import pytest
from numpy.testing import assert_array_equal, assert_equal

from brian2 import *
from brian2.core.network import schedule_propagation_offset
from brian2.devices.device import reinit_and_delete
from brian2.tests.utils import assert_allclose
from brian2.units.fundamentalunits import DIMENSIONLESS
from brian2.utils.logger import catch_logs


@pytest.mark.codegen_independent
def test_str_repr():
    """
    Test the string representation of a subgroup.
    """
    G = NeuronGroup(10, "v:1")
    SG = G[5:8]
    # very basic test, only make sure no error is raised
    assert len(str(SG))
    assert len(repr(SG))


def test_state_variables():
    """
    Test the setting and accessing of state variables in subgroups.
    """
    G = NeuronGroup(10, "v : volt")
    SG = G[4:9]
    with pytest.raises(DimensionMismatchError):
        SG.__setattr__("v", -70)
    SG.v_ = float(-80 * mV)
    assert_allclose(G.v, np.array([0, 0, 0, 0, -80, -80, -80, -80, -80, 0]) * mV)
    assert_allclose(SG.v, np.array([-80, -80, -80, -80, -80]) * mV)
    assert_allclose(
        G.v_, np.array([0, 0, 0, 0, -80, -80, -80, -80, -80, 0]) * float(mV)
    )
    assert_allclose(SG.v_, np.array([-80, -80, -80, -80, -80]) * float(mV))
    # You should also be able to set variables with a string
    SG.v = "v + i*mV"
    assert_allclose(SG.v[0], -80 * mV)
    assert_allclose(SG.v[4], -76 * mV)
    assert_allclose(G.v[4:9], -80 * mV + np.arange(5) * mV)

    # Calculating with state variables should work too
    assert all(G.v[4:9] - SG.v == 0)

    # And in-place modification should work as well
    SG.v += 10 * mV
    assert_allclose(G.v[4:9], -70 * mV + np.arange(5) * mV)
    SG.v *= 2
    assert_allclose(G.v[4:9], 2 * (-70 * mV + np.arange(5) * mV))
    # with unit checking
    with pytest.raises(DimensionMismatchError):
        SG.v.__iadd__(3 * second)
    with pytest.raises(DimensionMismatchError):
        SG.v.__iadd__(3)
    with pytest.raises(DimensionMismatchError):
        SG.v.__imul__(3 * second)

    # Indexing with subgroups
    assert_equal(G.v[SG], SG.v[:])


@pytest.mark.standalone_compatible
def test_state_variables_simple():
    G = NeuronGroup(
        10,
        """
        a : 1
        b : 1
        c : 1
        d : 1
        """,
    )
    SG = G[3:7]
    SG.a = 1
    SG.a["i == 0"] = 2
    SG.b = "i"
    SG.b["i == 3"] = "i * 2"
    SG.c = np.arange(3, 7)
    SG.d[1:2] = 4
    SG.d[2:4] = [1, 2]
    run(0 * ms)
    assert_equal(G.a[:], [0, 0, 0, 2, 1, 1, 1, 0, 0, 0])
    assert_equal(G.b[:], [0, 0, 0, 0, 1, 2, 6, 0, 0, 0])
    assert_equal(G.c[:], [0, 0, 0, 3, 4, 5, 6, 0, 0, 0])
    assert_equal(G.d[:], [0, 0, 0, 0, 4, 1, 2, 0, 0, 0])


def test_state_variables_string_indices():
    """
    Test accessing subgroups with string indices.
    """
    G = NeuronGroup(10, "v : volt")
    SG = G[4:9]
    assert len(SG.v["i>3"]) == 1

    G.v = np.arange(10) * mV
    assert len(SG.v["v>7.5*mV"]) == 1

    # Combined string indexing and assignment
    SG.v["i > 3"] = "i*10*mV"

    assert_allclose(G.v[:], [0, 1, 2, 3, 4, 5, 6, 7, 40, 9] * mV)


@pytest.mark.codegen_independent
def test_state_variables_group_as_index():
    G = NeuronGroup(10, "v : 1")
    SG = G[4:9]
    G.v[SG] = 1
    assert_equal(G.v[:], np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 0]))
    G.v = 1
    G.v[SG] = "2*v"
    assert_equal(G.v[:], np.array([1, 1, 1, 1, 2, 2, 2, 2, 2, 1]))


@pytest.mark.codegen_independent
def test_state_variables_group_as_index_problematic():
    G = NeuronGroup(10, "v : 1")
    SG = G[4:9]
    G.v = 1
    tests = [("i", 1), ("N", 1), ("N + i", 2), ("v", 0)]
    for value, n_warnings in tests:
        with catch_logs() as l:
            G.v.__setitem__(SG, value)
            assert (
                len(l) == n_warnings
            ), f"expected {int(n_warnings)}, got {len(l)} warnings"
            assert all(
                [entry[1].endswith("ambiguous_string_expression") for entry in l]
            )


@pytest.mark.standalone_compatible
def test_variableview_calculations():
    # Check that you can directly calculate with "variable views"
    G = NeuronGroup(
        10,
        """
        x : 1
        y : volt
        idx : integer
        """,
    )
    G.x = np.arange(10)
    G.y = np.arange(10)[::-1] * mV
    G.idx = np.arange(10, dtype=int)
    SG = G[3:8]

    assert_allclose(SG.x * SG.y, np.arange(3, 8) * np.arange(6, 1, -1) * mV)
    assert_allclose(-SG.x, -np.arange(3, 8))
    assert_allclose(-SG.y, -np.arange(6, 1, -1) * mV)

    assert_allclose(3 * SG.x, 3 * np.arange(3, 8))
    assert_allclose(3 * SG.y, 3 * np.arange(6, 1, -1) * mV)
    assert_allclose(SG.x * 3, 3 * np.arange(3, 8))
    assert_allclose(SG.y * 3, 3 * np.arange(6, 1, -1) * mV)
    assert_allclose(SG.x / 2.0, np.arange(3, 8) / 2.0)
    assert_allclose(SG.y / 2, np.arange(6, 1, -1) * mV / 2)
    assert_equal(SG.idx % 2, np.arange(3, 8, dtype=int) % 2)
    assert_allclose(SG.x + 2, 2 + np.arange(3, 8))
    assert_allclose(SG.y + 2 * mV, 2 * mV + np.arange(6, 1, -1) * mV)
    assert_allclose(2 + SG.x, 2 + np.arange(3, 8))
    assert_allclose(2 * mV + SG.y, 2 * mV + np.arange(6, 1, -1) * mV)
    assert_allclose(SG.x - 2, np.arange(3, 8) - 2)
    assert_allclose(SG.y - 2 * mV, np.arange(6, 1, -1) * mV - 2 * mV)
    assert_allclose(2 - SG.x, 2 - np.arange(3, 8))
    assert_allclose(2 * mV - SG.y, 2 * mV - np.arange(6, 1, -1) * mV)
    assert_allclose(SG.x**2, np.arange(3, 8) ** 2)
    assert_allclose(SG.y**2, (np.arange(6, 1, -1) * mV) ** 2)
    assert_allclose(2**SG.x, 2 ** np.arange(3, 8))

    # incorrect units
    with pytest.raises(DimensionMismatchError):
        SG.x + SG.y
    with pytest.raises(DimensionMismatchError):
        SG.x[:] + SG.y
    with pytest.raises(DimensionMismatchError):
        SG.x + SG.y[:]
    with pytest.raises(DimensionMismatchError):
        SG.x + 3 * mV
    with pytest.raises(DimensionMismatchError):
        3 * mV + SG.x
    with pytest.raises(DimensionMismatchError):
        SG.y + 3
    with pytest.raises(DimensionMismatchError):
        3 + SG.y
    with pytest.raises(TypeError):
        2**SG.y  # raising to a power with units


@pytest.mark.standalone_compatible
def test_variableview_properties():
    G = NeuronGroup(
        10,
        """
    x : 1
    y : volt
    idx : integer
    """,
    )
    # The below properties should not require access to the values
    G.x = "rand()"
    G.y = "rand()*mV"
    G.idx = "int(rand()*10)"
    SG = G[3:8]

    assert have_same_dimensions(SG.x.unit, DIMENSIONLESS)
    assert have_same_dimensions(SG.y.unit, volt)
    assert have_same_dimensions(SG.idx.unit, DIMENSIONLESS)
    # See github issue #1555
    assert SG.x.shape == SG.y.shape == SG.idx.shape == (5,)
    assert SG.x.ndim == SG.y.ndim == SG.idx.ndim == 1
    assert SG.x.dtype == SG.y.dtype == prefs.core.default_float_dtype
    assert SG.idx.dtype == np.int32


@pytest.mark.standalone_compatible
def test_state_monitor():
    G = NeuronGroup(10, "v : volt")
    G.v = np.arange(10) * volt
    SG = G[5:]
    mon_all = StateMonitor(SG, "v", record=True)
    mon_0 = StateMonitor(SG, "v", record=0)
    run(defaultclock.dt)

    assert_allclose(mon_0[0].v, mon_all[0].v)
    assert_allclose(mon_0[0].v, np.array([5]) * volt)
    assert_allclose(mon_all.v.flatten(), np.arange(5, 10) * volt)

    with pytest.raises(IndexError):
        mon_all[5]


def test_shared_variable():
    """Make sure that shared variables work with subgroups"""
    G = NeuronGroup(10, "v : volt (shared)")
    G.v = 1 * volt
    SG = G[5:]
    assert SG.v == 1 * volt


@pytest.mark.standalone_compatible
def test_synapse_creation():
    G1 = NeuronGroup(10, "")
    G2 = NeuronGroup(20, "")
    SG1 = G1[:5]
    SG2 = G2[10:]
    S = Synapses(SG1, SG2)
    S.connect(i=2, j=2)  # Should correspond to (2, 12)
    S.connect("i==2 and j==5")  # Should correspond to (2, 15)

    run(0 * ms)  # for standalone

    # Internally, the "real" neuron indices should be used
    assert_equal(S._synaptic_pre[:], np.array([2, 2]))
    assert_equal(S._synaptic_post[:], np.array([12, 15]))
    # For the user, the subgroup-relative indices should be presented
    assert_equal(S.i[:], np.array([2, 2]))
    assert_equal(S.j[:], np.array([2, 5]))
    # N_incoming and N_outgoing should also be correct
    assert all(S.N_outgoing[2, :] == 2)
    assert all(S.N_incoming[:, 2] == 1)
    assert all(S.N_incoming[:, 5] == 1)


@pytest.mark.standalone_compatible
def test_synapse_creation_state_vars():
    G1 = NeuronGroup(10, "v : 1")
    G2 = NeuronGroup(20, "v : 1")
    G1.v = "i"
    G2.v = "10 + i"
    SG1 = G1[:5]
    SG2 = G2[10:]

    # connect based on pre-/postsynaptic state variables
    S2 = Synapses(SG1, SG2, "w:1")
    S2.connect("v_pre > 2")

    S3 = Synapses(SG1, SG2, "w:1")
    S3.connect("v_post < 25")

    S4 = Synapses(SG2, SG1, "w:1")
    S4.connect("v_post > 2")

    S5 = Synapses(SG2, SG1, "w:1")
    S5.connect("v_pre < 25")

    run(0 * ms)  # for standalone

    assert len(S2) == 2 * len(SG2), str(len(S2))
    assert all(S2.v_pre[:] > 2)
    assert len(S3) == 5 * len(SG1), f"{len(S3)} != {5 * len(SG1)} "
    assert all(S3.v_post[:] < 25)

    assert len(S4) == 2 * len(SG2), str(len(S4))
    assert all(S4.v_post[:] > 2)
    assert len(S5) == 5 * len(SG1), f"{len(53)} != {5 * len(SG1)} "
    assert all(S5.v_pre[:] < 25)


@pytest.mark.standalone_compatible
def test_synapse_creation_generator():
    G1 = NeuronGroup(10, "v:1")
    G2 = NeuronGroup(20, "v:1")
    G1.v = "i"
    G2.v = "10 + i"
    SG1 = G1[:5]
    SG2 = G2[10:]
    S = Synapses(SG1, SG2, "w:1")
    S.connect(j="i*2 + k for k in range(2)")  # diverging connections

    # connect based on pre-/postsynaptic state variables
    S2 = Synapses(SG1, SG2, "w:1")
    S2.connect(j="k for k in range(N_post) if v_pre > 2")

    S3 = Synapses(SG1, SG2, "w:1")
    S3.connect(j="k for k in range(N_post) if v_post < 25")

    S4 = Synapses(SG2, SG1, "w:1")
    S4.connect(j="k for k in range(N_post) if v_post > 2")

    S5 = Synapses(SG2, SG1, "w:1")
    S5.connect(j="k for k in range(N_post) if v_pre < 25")

    run(0 * ms)  # for standalone

    # Internally, the "real" neuron indices should be used
    assert_equal(S._synaptic_pre[:], np.arange(5).repeat(2))
    assert_equal(S._synaptic_post[:], np.arange(10) + 10)
    # For the user, the subgroup-relative indices should be presented
    assert_equal(S.i[:], np.arange(5).repeat(2))
    assert_equal(S.j[:], np.arange(10))

    # N_incoming and N_outgoing should also be correct
    assert all(S.N_outgoing[:] == 2)
    assert all(S.N_incoming[:] == 1)

    assert len(S2) == 2 * len(SG2), str(len(S2))
    assert all(S2.v_pre[:] > 2)
    assert len(S3) == 5 * len(SG1), f"{len(S3)} != {5 * len(SG1)} "
    assert all(S3.v_post[:] < 25)

    assert len(S4) == 2 * len(SG2), str(len(S4))
    assert all(S4.v_post[:] > 2)
    assert len(S5) == 5 * len(SG1), f"{len(S5)} != {5 * len(SG1)} "
    assert all(S5.v_pre[:] < 25)


@pytest.mark.standalone_compatible
def test_synapse_creation_generator_multiple_synapses():
    G1 = NeuronGroup(10, "v:1")
    G2 = NeuronGroup(20, "v:1")
    G1.v = "i"
    G2.v = "10 + i"
    SG1 = G1[:5]
    SG2 = G2[10:]
    S1 = Synapses(SG1, SG2)
    S1.connect(j="k for k in range(N_post)", n="i")

    S2 = Synapses(SG1, SG2)
    S2.connect(j="k for k in range(N_post)", n="j")

    S3 = Synapses(SG2, SG1)
    S3.connect(j="k for k in range(N_post)", n="i")

    S4 = Synapses(SG2, SG1)
    S4.connect(j="k for k in range(N_post)", n="j")

    S5 = Synapses(SG1, SG2)
    S5.connect(j="k for k in range(N_post)", n="i+j")

    S6 = Synapses(SG2, SG1)
    S6.connect(j="k for k in range(N_post)", n="i+j")

    S7 = Synapses(SG1, SG2)
    S7.connect(j="k for k in range(N_post)", n="int(v_pre>2)*2")

    S8 = Synapses(SG2, SG1)
    S8.connect(j="k for k in range(N_post)", n="int(v_post>2)*2")

    S9 = Synapses(SG1, SG2)
    S9.connect(j="k for k in range(N_post)", n="int(v_post>22)*2")

    S10 = Synapses(SG2, SG1)
    S10.connect(j="k for k in range(N_post)", n="int(v_pre>22)*2")

    run(0 * ms)  # for standalone

    # straightforward loop instead of doing something clever...
    for source in range(len(SG1)):
        assert_equal(S1.j[source, :], np.arange(len(SG2)).repeat(source))
        assert_equal(S2.j[source, :], np.arange(len(SG2)).repeat(np.arange(len(SG2))))
        assert_equal(S3.i[:, source], np.arange(len(SG2)).repeat(np.arange(len(SG2))))
        assert_equal(S4.i[:, source], np.arange(len(SG2)).repeat(source))
        assert_equal(
            S5.j[source, :], np.arange(len(SG2)).repeat(np.arange(len(SG2)) + source)
        )
        assert_equal(
            S6.i[:, source], np.arange(len(SG2)).repeat(np.arange(len(SG2)) + source)
        )
        if source > 2:
            assert_equal(S7.j[source, :], np.arange(len(SG2)).repeat(2))
            assert_equal(S8.i[:, source], np.arange(len(SG2)).repeat(2))
        else:
            assert len(S7.j[source, :]) == 0
            assert len(S8.i[:, source]) == 0
        assert_equal(S9.j[source, :], np.arange(3, len(SG2)).repeat(2))
        assert_equal(S10.i[:, source], np.arange(3, len(SG2)).repeat(2))


@pytest.mark.standalone_compatible
def test_synapse_creation_generator_complex_ranges():
    G1 = NeuronGroup(10, "v:1")
    G2 = NeuronGroup(20, "v:1")
    G1.v = "i"
    G2.v = "10 + i"
    SG1 = G1[:5]
    SG2 = G2[10:]
    S = Synapses(SG1, SG2)
    S.connect(j="i+k for k in range(N_post-i)")  # Connect to all j>i

    # connect based on pre-/postsynaptic state variables
    S2 = Synapses(SG1, SG2)
    S2.connect(j="k for k in range(N_post * int(v_pre > 2))")

    # connect based on pre-/postsynaptic state variables
    S3 = Synapses(SG2, SG1)
    S3.connect(j="k for k in range(N_post * int(v_pre > 22))")

    run(0 * ms)  # for standalone

    for syn_source in range(5):
        # Internally, the "real" neuron indices should be used
        assert_equal(
            S._synaptic_post[syn_source, :],
            10 + syn_source + np.arange(10 - syn_source),
        )
        # For the user, the subgroup-relative indices should be presented
        assert_equal(S.j[syn_source, :], syn_source + np.arange(10 - syn_source))

    assert len(S2) == 2 * len(SG2), str(len(S2))
    assert all(S2.v_pre[:] > 2)
    assert len(S3) == 7 * len(SG1), str(len(S3))
    assert all(S3.v_pre[:] > 22)


@pytest.mark.standalone_compatible
def test_synapse_creation_generator_random():
    G1 = NeuronGroup(10, "v:1")
    G2 = NeuronGroup(20, "v:1")
    G1.v = "i"
    G2.v = "10 + i"
    SG1 = G1[:5]
    SG2 = G2[10:]

    # connect based on pre-/postsynaptic state variables
    S2 = Synapses(SG1, SG2)
    S2.connect(j="k for k in sample(N_post, p=1.0*int(v_pre > 2))")

    S3 = Synapses(SG2, SG1)
    S3.connect(j="k for k in sample(N_post, p=1.0*int(v_pre > 22))")

    run(0 * ms)  # for standalone

    assert len(S2) == 2 * len(SG2), str(len(S2))
    assert all(S2.v_pre[:] > 2)
    assert len(S3) == 7 * len(SG1), str(len(S3))
    assert all(S3.v_pre[:] > 22)


def test_synapse_access():
    G1 = NeuronGroup(10, "v:1")
    G1.v = "i"
    G2 = NeuronGroup(20, "v:1")
    G2.v = "i"
    SG1 = G1[:5]
    SG2 = G2[10:]
    S = Synapses(SG1, SG2, "w:1")
    S.connect(True)
    S.w["j == 0"] = 5
    assert all(S.w["j==0"] == 5)
    S.w[2, 2] = 7
    assert all(S.w["i==2 and j==2"] == 7)
    S.w = "2*j"
    assert all(S.w[:, 1] == 2)

    assert len(S.w[:, 10]) == 0
    assert len(S.w["j==10"]) == 0

    # Test referencing pre- and postsynaptic variables
    assert_equal(S.w[2:, :], S.w["v_pre >= 2"])
    assert_equal(S.w[:, :5], S.w["v_post < 15"])
    S.w = "v_post"
    assert_equal(S.w[:], S.j[:] + 10)
    S.w = "v_post + v_pre"
    assert_equal(S.w[:], S.j[:] + 10 + S.i[:])

    # Test using subgroups as indices
    assert len(S) == len(S.w[SG1, SG2])
    assert_equal(S.w[SG1, 1], S.w[:, 1])
    assert_equal(S.w[1, SG2], S.w[1, :])
    assert len(S.w[SG1, 10]) == 0


def test_synapses_access_subgroups():
    G1 = NeuronGroup(5, "x:1")
    G2 = NeuronGroup(10, "y:1")
    SG1 = G1[2:5]
    SG2 = G2[4:9]
    S = Synapses(G1, G2, "w:1")
    S.connect()
    S.w[SG1, SG2] = 1
    assert_equal(S.w["(i>=2 and i<5) and (j>=4 and j<9)"], 1)
    assert_equal(S.w["not ((i>=2 and i<5) and (j>=4 and j<9))"], 0)
    S.w = 0
    S.w[SG1, :] = 1
    assert_equal(S.w["i>=2 and i<5"], 1)
    assert_equal(S.w["not (i>=2 and i<5)"], 0)
    S.w = 0
    S.w[:, SG2] = 1
    assert_equal(S.w["j>=4 and j<9"], 1)
    assert_equal(S.w["not (j>=4 and j<9)"], 0)


@pytest.mark.codegen_independent
def test_synapses_access_subgroups_problematic():
    G1 = NeuronGroup(5, "x:1")
    G2 = NeuronGroup(10, "y:1")
    SG1 = G1[2:5]
    SG2 = G2[4:9]
    S = Synapses(G1, G2, "w:1")
    S.connect()

    # Note that "j" is not ambiguous, because the equivalent in the target group
    # is called "i" (this previously raised a warning)
    tests = [
        ((SG1, slice(None)), "i", 1),
        ((SG1, slice(None)), "i + N_pre", 2),
        ((SG1, slice(None)), "N_pre", 1),
        ((slice(None), SG2), "j", 0),
        ((slice(None), SG2), "N_post", 1),
        ((slice(None), SG2), "N", 1),
        ((SG1, SG2), "i", 1),
        ((SG1, SG2), "i + j", 1),
        ((SG1, SG2), "N_pre", 1),
        ((SG1, SG2), "j", 0),
        ((SG1, SG2), "N_post", 1),
        ((SG1, SG2), "N", 1),
        # These should not raise a warning
        ((SG1, SG2), "w", 0),
        ((SG1, SG2), "x_pre", 0),
        ((SG1, SG2), "y_post", 0),
        ((SG1, SG2), "y", 0),
    ]
    for item, value, n_warnings in tests:
        with catch_logs() as l:
            S.w.__setitem__(item, value)
            assert (
                len(l) == n_warnings
            ), f"expected {int(n_warnings)}, got {len(l)} warnings"
            assert all(
                [entry[1].endswith("ambiguous_string_expression") for entry in l]
            )


@pytest.mark.standalone_compatible
def test_subgroup_summed_variable():
    # Check in particular that only neurons targeted are reset to 0 (see github issue #925)
    source = NeuronGroup(1, "")
    target = NeuronGroup(5, "Iin : 1")
    target.Iin = 10
    target1 = target[1:2]
    target2 = target[3:]

    syn1 = Synapses(source, target1, "Iin_post = 5 : 1  (summed)")
    syn1.connect(True)
    syn2 = Synapses(source, target2, "Iin_post = 1 : 1  (summed)")
    syn2.connect(True)

    run(2 * defaultclock.dt)

    assert_array_equal(target.Iin, [10, 5, 10, 1, 1])


def test_subexpression_references():
    """
    Assure that subexpressions in targeted groups are handled correctly.
    """
    G = NeuronGroup(
        10,
        """
        v : 1
        v2 = 2*v : 1
        """,
    )
    G.v = np.arange(10)
    SG1 = G[:5]
    SG2 = G[5:]

    S1 = Synapses(
        SG1,
        SG2,
        """
        w : 1
        u = v2_post + 1 : 1
        x = v2_pre + 1 : 1
        """,
    )
    S1.connect("i==(5-1-j)")
    assert_equal(S1.i[:], np.arange(5))
    assert_equal(S1.j[:], np.arange(5)[::-1])
    assert_equal(S1.u[:], np.arange(10)[:-6:-1] * 2 + 1)
    assert_equal(S1.x[:], np.arange(5) * 2 + 1)

    S2 = Synapses(
        G,
        SG2,
        """
        w : 1
        u = v2_post + 1 : 1
        x = v2_pre + 1 : 1
        """,
    )
    S2.connect("i==(5-1-j)")
    assert_equal(S2.i[:], np.arange(5))
    assert_equal(S2.j[:], np.arange(5)[::-1])
    assert_equal(S2.u[:], np.arange(10)[:-6:-1] * 2 + 1)
    assert_equal(S2.x[:], np.arange(5) * 2 + 1)

    S3 = Synapses(
        SG1,
        G,
        """
        w : 1
        u = v2_post + 1 : 1
        x = v2_pre + 1 : 1
        """,
    )
    S3.connect("i==(10-1-j)")
    assert_equal(S3.i[:], np.arange(5))
    assert_equal(S3.j[:], np.arange(10)[:-6:-1])
    assert_equal(S3.u[:], np.arange(10)[:-6:-1] * 2 + 1)
    assert_equal(S3.x[:], np.arange(5) * 2 + 1)


def test_subexpression_no_references():
    """
    Assure that subexpressions  are handled correctly, even
    when the subgroups are created on-the-fly.
    """
    G = NeuronGroup(
        10,
        """
        v : 1
        v2 = 2*v : 1
        """,
    )
    G.v = np.arange(10)

    assert_equal(G[5:].v2, np.arange(5, 10) * 2)

    S1 = Synapses(
        G[:5],
        G[5:],
        """
        w : 1
        u = v2_post + 1 : 1
        x = v2_pre + 1 : 1
        """,
    )
    S1.connect("i==(5-1-j)")
    assert_equal(S1.i[:], np.arange(5))
    assert_equal(S1.j[:], np.arange(5)[::-1])
    assert_equal(S1.u[:], np.arange(10)[:-6:-1] * 2 + 1)
    assert_equal(S1.x[:], np.arange(5) * 2 + 1)

    S2 = Synapses(
        G,
        G[5:],
        """
        w : 1
        u = v2_post + 1 : 1
        x = v2_pre + 1 : 1
        """,
    )
    S2.connect("i==(5-1-j)")
    assert_equal(S2.i[:], np.arange(5))
    assert_equal(S2.j[:], np.arange(5)[::-1])
    assert_equal(S2.u[:], np.arange(10)[:-6:-1] * 2 + 1)
    assert_equal(S2.x[:], np.arange(5) * 2 + 1)

    S3 = Synapses(
        G[:5],
        G,
        """
        w : 1
        u = v2_post + 1 : 1
        x = v2_pre + 1 : 1
        """,
    )
    S3.connect("i==(10-1-j)")
    assert_equal(S3.i[:], np.arange(5))
    assert_equal(S3.j[:], np.arange(10)[:-6:-1])
    assert_equal(S3.u[:], np.arange(10)[:-6:-1] * 2 + 1)
    assert_equal(S3.x[:], np.arange(5) * 2 + 1)


@pytest.mark.standalone_compatible
def test_synaptic_propagation():
    G1 = NeuronGroup(10, "v:1", threshold="v>1", reset="v=0")
    G1.v["i%2==1"] = 1.1  # odd numbers should spike
    G2 = NeuronGroup(20, "v:1")
    SG1 = G1[1:6]
    SG2 = G2[10:]
    S = Synapses(SG1, SG2, on_pre="v+=1")
    S.connect("i==j")
    run(defaultclock.dt + schedule_propagation_offset())
    expected = np.zeros(len(G2))
    # Neurons 1, 3, 5 spiked and are connected to 10, 12, 14
    expected[[10, 12, 14]] = 1
    assert_equal(np.asarray(G2.v).flatten(), expected)


@pytest.mark.standalone_compatible
def test_synaptic_propagation_2():
    # This tests for the bug in github issue #461
    source = NeuronGroup(100, "", threshold="True")
    sub_source = source[99:]
    target = NeuronGroup(1, "v:1")
    syn = Synapses(sub_source, target, on_pre="v+=1")
    syn.connect()
    run(defaultclock.dt + schedule_propagation_offset())
    assert target.v[0] == 1.0


@pytest.mark.standalone_compatible
def test_run_regularly():
    # See github issue #922

    group = NeuronGroup(10, "v: integer")
    # Full group
    group.run_regularly("v += 16")
    # Subgroup with explicit reference
    subgroup = group[:2]
    subgroup.run_regularly("v += 8")
    # Subgroup with explicit reference and reference for run_regularly operation
    subgroup2 = group[2:4]
    updater = subgroup2.run_regularly("v += 4")
    # Subgroup without reference
    group[4:6].run_regularly("v += 2")
    # Subgroup without reference, with reference for run_regularly operation
    updater2 = group[6:8].run_regularly("v += 1")

    run(defaultclock.dt)
    assert_array_equal(group.v, [24, 24, 20, 20, 18, 18, 17, 17, 16, 16])


@pytest.mark.standalone_compatible
def test_spike_monitor():
    G = NeuronGroup(10, "v:1", threshold="v>1", reset="v=0")
    G.v[0] = 1.1
    G.v[2] = 1.1
    G.v[5] = 1.1
    SG = G[3:]
    SG2 = G[:3]
    s_mon = SpikeMonitor(G)
    sub_s_mon = SpikeMonitor(SG)
    sub_s_mon2 = SpikeMonitor(SG2)
    run(defaultclock.dt)
    assert_equal(s_mon.i, np.array([0, 2, 5]))
    assert_equal(s_mon.t_, np.zeros(3))
    assert_equal(sub_s_mon.i, np.array([2]))
    assert_equal(sub_s_mon.t_, np.zeros(1))
    assert_equal(sub_s_mon2.i, np.array([0, 2]))
    assert_equal(sub_s_mon2.t_, np.zeros(2))
    expected = np.zeros(10, dtype=int)
    expected[[0, 2, 5]] = 1
    assert_equal(s_mon.count, expected)
    expected = np.zeros(7, dtype=int)
    expected[[2]] = 1
    assert_equal(sub_s_mon.count, expected)
    assert_equal(sub_s_mon2.count, np.array([1, 0, 1]))


@pytest.mark.codegen_independent
def test_wrong_indexing():
    G = NeuronGroup(10, "v:1")
    with pytest.raises(TypeError):
        G["string"]

    with pytest.raises(IndexError):
        G[10]
    with pytest.raises(IndexError):
        G[10:]
    with pytest.raises(IndexError):
        G[::2]
    with pytest.raises(IndexError):
        G[3:2]
    with pytest.raises(IndexError):
        G[[5, 4, 3]]
    with pytest.raises(IndexError):
        G[[2, 4, 6]]
    with pytest.raises(IndexError):
        G[[-1, 0, 1]]
    with pytest.raises(IndexError):
        G[[9, 10, 11]]
    with pytest.raises(IndexError):
        G[[9, 10]]
    with pytest.raises(IndexError):
        G[[10, 11]]
    with pytest.raises(TypeError):
        G[[2.5, 3.5, 4.5]]


@pytest.mark.codegen_independent
def test_alternative_indexing():
    G = NeuronGroup(10, "v : integer")
    G.v = "i"
    assert_equal(G[-3:].v, np.array([7, 8, 9]))
    assert_equal(G[3].v, np.array([3]))
    assert_equal(G[[3, 4, 5]].v, np.array([3, 4, 5]))


def test_no_reference_1():
    """
    Using subgroups without keeping an explicit reference. Basic access.
    """
    G = NeuronGroup(10, "v:1")
    G.v = np.arange(10)
    assert_equal(G[:5].v[:], G.v[:5])


@pytest.mark.standalone_compatible
def test_no_reference_2():
    """
    Using subgroups without keeping an explicit reference. Monitors
    """
    G = NeuronGroup(2, "v:1", threshold="v>1", reset="v=0")
    G.v = [0, 1.1]
    state_mon = StateMonitor(G[:1], "v", record=True)
    spike_mon = SpikeMonitor(G[1:])
    rate_mon = PopulationRateMonitor(G[:2])
    run(2 * defaultclock.dt)
    assert_equal(state_mon[0].v[:], np.zeros(2))
    assert_equal(spike_mon.i[:], np.array([0]))
    assert_equal(spike_mon.t[:], np.array([0]) * second)
    assert_equal(rate_mon.rate[:], np.array([0.5, 0]) / defaultclock.dt)


@pytest.mark.standalone_compatible
def test_no_reference_3():
    """
    Using subgroups without keeping an explicit reference. Monitors
    """
    G = NeuronGroup(2, "v:1", threshold="v>1", reset="v=0")
    G.v = [1.1, 0]
    S = Synapses(G[:1], G[1:], on_pre="v+=1")
    S.connect()
    run(defaultclock.dt + schedule_propagation_offset())
    assert_equal(G.v[:], np.array([0, 1]))


@pytest.mark.standalone_compatible
def test_no_reference_4():
    """
    Using subgroups without keeping an explicit reference. Synapses
    """
    G1 = NeuronGroup(10, "v:1", threshold="v>1", reset="v=0")
    G1.v["i%2==1"] = 1.1  # odd numbers should spike
    G2 = NeuronGroup(20, "v:1")
    S = Synapses(G1[1:6], G2[10:], on_pre="v+=1")
    S.connect("i==j")
    run(defaultclock.dt + schedule_propagation_offset())
    expected = np.zeros(len(G2))
    # Neurons 1, 3, 5 spiked and are connected to 10, 12, 14
    expected[[10, 12, 14]] = 1
    assert_equal(np.asarray(G2.v).flatten(), expected)


def test_recursive_subgroup():
    """
    Create a subgroup of a subgroup
    """
    G = NeuronGroup(10, "v : 1")
    G.v = "i"
    SG = G[3:8]
    SG2 = SG[2:4]
    assert_equal(SG2.v[:], np.array([5, 6]))
    assert_equal(SG2.v[:], SG.v[2:4])
    assert SG2.source.name == G.name


if __name__ == "__main__":
    test_str_repr()
    test_state_variables()
    test_state_variables_simple()
    test_state_variables_string_indices()
    test_state_variables_group_as_index()
    test_state_variables_group_as_index_problematic()
    test_state_monitor()
    test_shared_variable()
    test_synapse_creation()
    test_synapse_creation_state_vars()
    test_synapse_creation_generator()
    test_synapse_creation_generator_complex_ranges()
    test_synapse_creation_generator_random()
    test_synapse_creation_generator_multiple_synapses()
    test_synapse_access()
    test_synapses_access_subgroups()
    test_synapses_access_subgroups_problematic()
    test_subgroup_summed_variable()
    test_subexpression_references()
    test_subexpression_no_references()
    test_synaptic_propagation()
    test_synaptic_propagation_2()
    test_run_regularly()
    test_spike_monitor()
    test_wrong_indexing()
    test_no_reference_1()
    test_no_reference_2()
    test_no_reference_3()
    test_no_reference_4()
    test_recursive_subgroup()



"""
Parallel processes using standalone mode

This example use multiprocessing to run several simulations in parallel.
The code is using the C++ standalone mode to compile and execute the code.

The generated code is stored in a ``standalone{pid}`` directory, with ``pid``
being the id of each process.

Note that the `set_device` call should be in the ``run_sim`` function.

By moving the `set_device` line into the parallelised function, it creates one
C++ standalone device per process.
The ``device.reinit()`` needs to be called` if you are running multiple
simulations per process (there are 10 tau values and num_proc = 4).

Each simulation uses it's own code folder to generate the code for the
simulation, controlled by the directory keyword to the set_device
call. By setting ``directory=None``, a temporary folder with random name is
created. This way, each simulation uses a different folder for code generation
and there is nothing shared between the parallel processes.

If you don't set the directory argument, it defaults to ``directory="output"``.
In that case each process would use the same files to try to generate and
compile your simulation, which would lead to compile/execution errors.

Setting ``directory=f"standalone{pid}"`` is even better than using
``directory=None`` in this case. That is, giving each parallel process
*it's own directory to work on*. This way you avoid the problem of multiple
processes working on the same code directories. But you also don't need to
recompile the entire project at each simulation. What happens is that in the
generated code in two consecutive simulations in a single process will only
differ slightly (in this case only the tau parameter). The compiler will
therefore only recompile the file that has changed and not the entire project.

The ``numb_proc`` sets the number of processes. ``run_sim`` is just a toy
example that creates a single neuron and connects a `StateMonitor` to record
the voltage.

For more details see the `discussion in the Brian forum <https://brian.discourse.group/t/multiprocessing-in-standalone-mode/142/2>`_.

Note that Python's `multiprocessing` module cannot deal with user-defined functions (including `TimedArray`) and other
complex code structures. If you run into `PicklingError` or `AttributeError` exceptions, you might
have to use the `pathos` (https://pypi.org/project/pathos) package instead, which can handle more complex
code structures.
"""
import os
import multiprocessing
from time import time as wall_time
from os import system
from brian2 import *

def run_sim(tau):
    pid = os.getpid()
    directory = f"standalone{pid}"
    set_device('cpp_standalone', directory=directory)
    print(f'RUNNING {pid}')

    G = NeuronGroup(1, 'dv/dt = -v/tau : 1', method='euler')
    G.v = 1

    mon = StateMonitor(G, 'v', record=0)
    net = Network()
    net.add(G, mon)
    net.run(100 * ms)
    res = (mon.t/ms, mon.v[0])

    device.reinit()

    print(f'FINISHED {pid}')
    return res


if __name__ == "__main__":
    start_time = wall_time()
    
    num_proc = 4
    tau_values = np.arange(10)*ms + 5*ms
    with multiprocessing.Pool(num_proc) as p:
        results = p.map(run_sim, tau_values)

    print(f"Done in {wall_time() - start_time:10.3f}")

    for tau_value, (t, v) in zip(tau_values, results):
        plt.plot(t, v, label=str(tau_value))
    plt.legend()
    plt.show()



"""
Runtime targets for code generation.
"""

# Register the base category before importing the indivial codegen targets with
# their subcategories

from brian2.core.preferences import prefs
from brian2.utils.logger import get_logger

prefs.register_preferences(
    "codegen.runtime",
    "Runtime codegen preferences (see subcategories for individual targets)",
)

logger = get_logger(__name__)

from .numpy_rt import *

try:
    from .cython_rt import *
except ImportError:
    pass  # todo: raise a warning?
try:
    from .GSLcython_rt import *
except ImportError:
    pass



"""
Module providing `DictImportExport` and `PandasImportExport` (requiring a
working installation of pandas).
"""

import numpy as np

from .importexport import ImportExport


class DictImportExport(ImportExport):
    """
    An importer/exporter for variables in format of dict of numpy arrays.
    """

    @property
    def name(self):
        return "dict"

    @staticmethod
    def export_data(group, variables, units=True, level=0):
        data = {}
        for var in variables:
            data[var] = np.array(
                group.state(var, use_units=units, level=level + 1),
                copy=True,
                subok=True,
            )
        return data

    @staticmethod
    def import_data(group, data, units=True, level=0):
        for key, value in data.items():
            if group.variables[key].read_only:
                raise TypeError(f"Variable {key} is read-only.")
            group.state(key, use_units=units, level=level + 1)[:] = value


class PandasImportExport(ImportExport):
    """
    An importer/exporter for variables in pandas DataFrame format.
    """

    @property
    def name(self):
        return "pandas"

    @staticmethod
    def export_data(group, variables, units=True, level=0):
        # pandas is not a default brian2 dependency, only import it here
        try:
            import pandas as pd
        except ImportError as ex:
            raise ImportError(
                "Exporting to pandas needs a working installation"
                " of pandas. Importing pandas failed: " + str(ex)
            )
        if units:
            raise NotImplementedError(
                "Units not supported when exporting to pandas data frame"
            )
        # we take advantage of the already implemented exporter
        data = DictImportExport.export_data(group, variables, units=units, level=level)
        pandas_data = pd.DataFrame(data)
        return pandas_data

    @staticmethod
    def import_data(group, data, units=True, level=0):
        if units:
            raise NotImplementedError(
                "Units not supported when importing from pandas data frame"
            )
        colnames = data.columns
        array_data = data.values
        for e, colname in enumerate(colnames):
            if group.variables[colname].read_only:
                raise TypeError(f"Variable '{colname}' is read-only.")
            state = group.state(colname, use_units=units, level=level + 1)
            state[:] = np.squeeze(array_data[:, e])



"""
Some basic tests for the `Variable` system
"""

from collections import namedtuple

import numpy as np
import pytest

from brian2.core.preferences import prefs
from brian2.core.variables import *
from brian2.units.allunits import second
from brian2.units.fundamentalunits import Unit


@pytest.mark.codegen_independent
def test_construction_errors():
    # Boolean variable that isn't dimensionless
    with pytest.raises(ValueError):
        Variable(name="name", dimensions=second.dim, dtype=bool)

    # Dynamic array variable that is constant but not constant in size
    with pytest.raises(ValueError):
        DynamicArrayVariable(
            name="name",
            owner=None,
            size=0,
            device=None,
            constant=True,
            needs_reference_update=True,
        )


@pytest.mark.codegen_independent
def test_str_repr():
    # Basic test that the str/repr methods work
    FakeGroup = namedtuple("G", ["name"])
    group = FakeGroup(name="groupname")
    variables = [
        Variable(name="name", dimensions=second.dim),
        Constant(name="name", dimensions=second.dim, value=1.0),
        AuxiliaryVariable(name="name", dimensions=second.dim),
        ArrayVariable(
            name="name", dimensions=second.dim, owner=None, size=10, device=None
        ),
        DynamicArrayVariable(
            name="name", dimensions=second.dim, owner=None, size=0, device=None
        ),
        Subexpression(
            name="sub", dimensions=second.dim, expr="a+b", owner=group, device=None
        ),
    ]
    for var in variables:
        assert len(str(var))
        # The repr value should contain the name of the class
        assert len(repr(var)) and var.__class__.__name__ in repr(var)


@pytest.mark.codegen_independent
def test_dtype_str():
    FakeGroup = namedtuple("G", ["name"])
    group = FakeGroup(name="groupname")
    for d in ["int32", "int64", "float32", "float64", "bool", "int", "float"]:
        nd = np.dtype(d)
        for var in [
            Constant(name="name", value=np.zeros(1, dtype=nd)[0]),
            AuxiliaryVariable(name="name", dtype=nd),
            ArrayVariable(name="name", owner=None, size=10, device=None, dtype=nd),
            DynamicArrayVariable(
                name="name", owner=None, dtype=nd, size=0, device=None
            ),
            Subexpression(name="sub", expr="a+b", owner=group, device=None, dtype=nd),
        ]:
            assert var.dtype_str.startswith(d)


if __name__ == "__main__":
    test_construction_errors()
    test_str_repr()
    test_dtype_str()



#!/usr/bin/env python
"""
The most simple case how to use standalone mode with several `run` calls.
"""
from brian2 import *
set_device('cpp_standalone', build_on_run=False)

tau = 10*ms
I = 1  # input current
eqs = '''
dv/dt = (I-v)/tau : 1
'''
G = NeuronGroup(10, eqs, method='exact')
G.v = 'rand()'
mon = StateMonitor(G, 'v', record=True)
run(20*ms)
I = 0
run(80*ms)
# Actually generate/compile/run the code:
device.build()

plt.plot(mon.t/ms, mon.v.T)
plt.gca().set(xlabel='t (ms)', ylabel='v')
plt.show()



from io import StringIO

import pytest
from numpy import float32, float64
from numpy.testing import assert_equal

from brian2 import amp, restore_initial_state, volt
from brian2.core.preferences import (
    BrianGlobalPreferences,
    BrianGlobalPreferencesView,
    BrianPreference,
    DefaultValidator,
    PreferenceError,
)


@pytest.mark.codegen_independent
def test_defaultvalidator():
    # Test that the default validator checks the class
    validator = DefaultValidator(5)
    assert validator(3)
    assert not validator("3")
    validator = DefaultValidator("astring")
    assert validator("another")
    assert not validator(3)
    # test that the default validator checks the units
    validator = DefaultValidator(3 * volt)
    assert validator(2 * volt)
    assert not validator(1 * amp)


@pytest.mark.codegen_independent
def test_brianpreference():
    # check default args
    pref = BrianPreference(1.0 / 3, "docs")
    assert not pref.validator(1)
    assert pref.docs == "docs"
    assert pref.default == 1.0 / 3
    assert pref.representor(pref.default) == repr(1.0 / 3)


@pytest.mark.codegen_independent
def test_preference_name_checking():
    """
    Test that you cannot set illegal preference names.
    """
    gp = BrianGlobalPreferences()

    # Name that starts with an underscore
    with pytest.raises(PreferenceError):
        gp.register_preferences(
            "dummy",
            "dummy doc",
            _notalegalname=BrianPreference(True, "some preference"),
        )

    # Name that clashes with a method name
    with pytest.raises(PreferenceError):
        gp.register_preferences(
            "dummy", "dummy doc", update=BrianPreference(True, "some preference")
        )

    gp.register_preferences(
        "a", "dummy doc", b=BrianPreference(True, "some preference")
    )

    # Trying to register a subcategory that would shadow a preference
    with pytest.raises(PreferenceError):
        gp.register_preferences(
            "a.b", "dummy doc", name=BrianPreference(True, "some preference")
        )

    gp.register_preferences(
        "b.c", "dummy doc", name=BrianPreference(True, "some preference")
    )

    # Trying to register a preference that clashes with an existing category
    with pytest.raises(PreferenceError):
        gp.register_preferences(
            "b", "dummy doc", c=BrianPreference(True, "some preference")
        )


@pytest.mark.codegen_independent
def test_brianglobalpreferences():
    # test that pre-setting a nonexistent preference in a subsequently
    # existing base name raises an error at the correct point
    gp = BrianGlobalPreferences()

    # This shouldn't work, in user code only registered preferences can be set
    with pytest.raises(PreferenceError):
        gp.__setitem__("a.b", 5)

    # This uses the method that is used when reading preferences from a file
    gp._set_preference("a.b", 5)
    gp._set_preference("a.c", 5)
    with pytest.raises(PreferenceError):
        gp.register_preferences("a", "docs for a", b=BrianPreference(5, "docs for b"))
    # test that post-setting a nonexistent preference in an existing base
    # name raises an error
    gp = BrianGlobalPreferences()
    gp.register_preferences("a", "docs for a", b=BrianPreference(5, "docs for b"))
    with pytest.raises(PreferenceError):
        gp.__setitem__("a.c", 5)
    # Test pre and post-setting some correct names but valid and invalid values
    gp = BrianGlobalPreferences()
    gp._set_preference("a.b", 5)
    gp.register_preferences(
        "a",
        "docs for a",
        b=BrianPreference(5, "docs for b"),
        c=BrianPreference(1 * volt, "docs for c"),
        d=BrianPreference(0, "docs for d", validator=lambda x: x >= 0),
        e=BrianPreference(float64, "docs for e", representor=lambda x: x.__name__),
    )
    assert gp["a.c"] == 1 * volt
    gp["a.c"] = 2 * volt
    with pytest.raises(PreferenceError):
        gp.__setitem__("a.c", 3 * amp)
    gp["a.d"] = 2.0
    with pytest.raises(PreferenceError):
        gp.__setitem__("a.d", -1)
    gp["a.e"] = float32
    with pytest.raises(PreferenceError):
        gp.__setitem__("a.e", 0)
    # test backup and restore
    gp._backup()
    gp["a.d"] = 10
    assert gp["a.d"] == 10
    gp._restore()
    assert gp["a.d"] == 2.0
    # test that documentation and as_file generation runs without error, but
    # don't test for values because we might change the organisation of it
    assert len(gp.get_documentation())
    gp.as_file
    gp.defaults_as_file
    # test that reading a preference file works as expected
    pref_file = StringIO(
        """
        # a comment
        a.b = 10
        [a]
        c = 5*volt
        d = 1
        e = float64
        """
    )
    gp.read_preference_file(pref_file)
    assert gp["a.b"] == 10
    assert gp["a.c"] == 5 * volt
    assert gp["a.d"] == 1
    assert gp["a.e"] == float64
    # test that reading a badly formatted prefs file fails
    pref_file = StringIO(
        """
        [a
        b = 10
        """
    )
    with pytest.raises(PreferenceError):
        gp.read_preference_file(pref_file)
    # test that reading a well formatted prefs file with an invalid value fails
    pref_file = StringIO(
        """
        a.b = 'oh no, not a string'
        """
    )
    with pytest.raises(PreferenceError):
        gp.read_preference_file(pref_file)
    # assert that writing the prefs to a file and loading them gives the
    # same values
    gp = BrianGlobalPreferences()
    gp.register_preferences(
        "a",
        "docs for a",
        b=BrianPreference(5, "docs for b"),
    )
    gp._backup()
    gp["a.b"] = 10
    str_modified = gp.as_file
    str_defaults = gp.defaults_as_file
    gp["a.b"] = 15
    gp.read_preference_file(StringIO(str_modified))
    assert gp["a.b"] == 10
    gp.read_preference_file(StringIO(str_defaults))
    assert gp["a.b"] == 5
    # check that load_preferences works, but nothing about its values
    gp = BrianGlobalPreferences()
    gp.load_preferences()

    # Check that resetting to default preferences works
    gp = BrianGlobalPreferences()
    gp.register_preferences("a", "docs for a", b=BrianPreference(5, "docs for b"))
    assert gp["a.b"] == 5
    gp["a.b"] = 7
    assert gp["a.b"] == 7
    gp.reset_to_defaults()
    assert gp["a.b"] == 5


@pytest.mark.codegen_independent
def test_preference_name_access():
    """
    Test various ways of accessing preferences
    """

    gp = BrianGlobalPreferences()

    gp.register_preferences(
        "main", "main category", name=BrianPreference(True, "some preference")
    )
    gp.register_preferences(
        "main.sub", "subcategory", name2=BrianPreference(True, "some preference")
    )

    gp.register_preferences("main.sub_no_pref", "subcategory without preference")
    gp.register_preferences(
        "main.sub_no_pref.sub",
        "deep subcategory",
        name=BrianPreference(True, "some preference"),
    )

    # Keyword based access
    assert gp["main.name"]
    assert gp["main.sub.name2"]
    assert gp["main.sub_no_pref.sub.name"]
    gp["main.name"] = False
    gp["main.sub.name2"] = False
    gp["main.sub_no_pref.sub.name"] = False

    # Attribute based access
    assert not gp.main.name  # we set it to False above
    assert not gp.main.sub.name2
    assert not gp.main.sub_no_pref.sub.name
    gp.main.name = True
    gp.main.sub.name2 = True
    gp.main.sub_no_pref.sub.name = True

    # Mixed access
    assert gp.main["name"]
    assert gp["main"].name
    assert gp.main["sub"].name2
    assert gp["main"].sub["name2"]

    # Accessing categories
    assert isinstance(gp["main"], BrianGlobalPreferencesView)
    assert isinstance(gp["main.sub"], BrianGlobalPreferencesView)
    assert isinstance(gp.main, BrianGlobalPreferencesView)
    assert isinstance(gp.main.sub, BrianGlobalPreferencesView)

    # Setting categories shouldn't work
    with pytest.raises(PreferenceError):
        gp.__setitem__("main", None)
    with pytest.raises(PreferenceError):
        gp.__setattr__("main", None)
    with pytest.raises(PreferenceError):
        gp.main.__setitem__("sub", None)
    with pytest.raises(PreferenceError):
        gp.main.__setattr__("sub", None)

    # Neither should deleting categories or preferences
    with pytest.raises(PreferenceError):
        gp.__delitem__("main")
    with pytest.raises(PreferenceError):
        gp.__delattr__("main")
    with pytest.raises(PreferenceError):
        gp.main.__delitem__("name")
    with pytest.raises(PreferenceError):
        gp.main.__delattr__("name")
    with pytest.raises(PreferenceError):
        gp.main.__delitem__("sub")
    with pytest.raises(PreferenceError):
        gp.main.__delattr__("sub")

    # Errors for accessing non-existing preferences
    with pytest.raises(KeyError):
        gp["main.doesnotexist"]
    with pytest.raises(KeyError):
        gp["nonexisting.name"]
    # Note that it is important to raise AttributeError here, since some tools like
    # Google Collab check for a `.shape` attribute to display debugging information
    # for numpy arrays. More generally, a call like `hasattr(gp, "main.doesnotexist")`
    # would fail with an error instead of returning `False`
    with pytest.raises(AttributeError):
        gp.main.doesnotexist
    with pytest.raises(AttributeError):
        gp.nonexisting.name

    # Check dictionary functionality
    for name, value in gp.items():
        assert gp[name] == value

    for name, value in gp.main.items():
        assert gp.main[name] == value

    assert len(gp) == 3  # three preferences in total
    assert len(gp["main"]) == 3  # all preferences are in the main category
    assert len(gp["main.sub"]) == 1  # one preference in main.sub

    assert "main.name" in gp
    assert "name" in gp["main"]
    assert "name2" in gp["main.sub"]
    assert not "name" in gp["main.sub"]

    gp["main.name"] = True
    gp.update({"main.name": False})
    assert not gp["main.name"]

    gp.main.update({"name": True})
    assert gp["main.name"]

    # Class based functionality
    assert "main" in dir(gp)
    assert "sub" in dir(gp.main)
    assert "name" in dir(gp.main)

    # Check that the fiddling with getattr and setattr did not destroy the
    # access to standard attributes
    assert len(gp.prefs)
    assert gp.main._basename == "main"

    # Check that acessing internal attributes work
    assert len(gp.prefs.__doc__)


@pytest.mark.codegen_independent
def test_str_repr():
    # Just test whether str and repr do not throw an error and return something
    gp = BrianGlobalPreferences()
    gp.register_preferences(
        "main", "main category", name=BrianPreference(True, "some preference")
    )

    assert len(str(gp))
    assert len(repr(gp))
    assert len(str(gp.main))
    assert len(repr(gp.main))


if __name__ == "__main__":
    for t in [
        test_defaultvalidator,
        test_brianpreference,
        test_brianglobalpreferences,
        test_preference_name_checking,
        test_preference_name_access,
    ]:
        t()
        restore_initial_state()



"""
Utility functions to get information about the environment Brian is running in.
"""

import builtins


def running_from_ipython():
    """
    Check whether we are currently running under ipython.

    Returns
    -------
    ipython : bool
        Whether running under ipython or not.
    """
    return getattr(builtins, "__IPYTHON__", False)



import copy
import logging
import os
import tempfile
import uuid
import weakref

import numpy as np
import pytest
from numpy.testing import assert_array_equal, assert_equal

from brian2 import (
    BrianLogger,
    BrianObject,
    Clock,
    Hz,
    MagicError,
    MagicNetwork,
    Network,
    NetworkOperation,
    NeuronGroup,
    PoissonGroup,
    PopulationRateMonitor,
    Quantity,
    SpikeGeneratorGroup,
    SpikeMonitor,
    StateMonitor,
    Synapses,
    TimedArray,
    collect,
    defaultclock,
    magic_network,
    ms,
    network_operation,
    prefs,
    profiling_summary,
    restore,
    run,
    second,
    start_scope,
    stop,
    store,
    us,
)
from brian2.core.network import schedule_propagation_offset, scheduling_summary
from brian2.devices.device import (
    Device,
    RuntimeDevice,
    all_devices,
    device,
    get_device,
    reinit_and_delete,
    reset_device,
    set_device,
)
from brian2.tests.utils import assert_allclose
from brian2.utils.logger import catch_logs


@pytest.mark.codegen_independent
def test_incorrect_network_use():
    """Test some wrong uses of `Network` and `MagicNetwork`"""
    with pytest.raises(TypeError):
        Network(name="mynet", anotherkwd="does not exist")
    with pytest.raises(TypeError):
        Network("not a BrianObject")
    net = Network()
    with pytest.raises(TypeError):
        net.add("not a BrianObject")
    with pytest.raises(ValueError):
        MagicNetwork()
    G = NeuronGroup(10, "v:1")
    net.add(G)
    with pytest.raises(TypeError):
        net.remove(object())
    with pytest.raises(MagicError):
        magic_network.add(G)
    with pytest.raises(MagicError):
        magic_network.remove(G)


@pytest.mark.codegen_independent
def test_network_contains():
    """
    Test `Network.__contains__`.
    """
    G = NeuronGroup(1, "v:1", name="mygroup")
    net = Network(G)
    assert "mygroup" in net
    assert "neurongroup" not in net


@pytest.mark.codegen_independent
def test_empty_network():
    # Check that an empty network functions correctly
    net = Network()
    net.run(1 * second)


class Counter(BrianObject):
    add_to_magic_network = True

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.count = 0
        self.state = {"state": 0}

    def get_states(self, *args, **kwds):
        return dict(self.state)

    def set_states(self, values, *args, **kwds):
        for k, v in values.items():
            self.state[k] = v

    def run(self):
        self.count += 1


class CounterWithContained(Counter):
    add_to_magic_network = True

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.sub_counter = Counter()
        self.contained_objects.append(self.sub_counter)


@pytest.mark.codegen_independent
def test_network_single_object():
    # Check that a network with a single object functions correctly
    x = Counter()
    net = Network(x)
    net.run(1 * ms)
    assert_equal(x.count, 10)


@pytest.mark.codegen_independent
def test_network_two_objects():
    # Check that a network with two objects and the same clock function correctly
    x = Counter(order=5)
    y = Counter(order=6)
    net = Network()
    net.add([x, [y]])  # check that a funky way of adding objects work correctly
    net.run(1 * ms)
    assert_equal(len(net.objects), 2)
    assert_equal(x.count, 10)
    assert_equal(y.count, 10)


@pytest.mark.codegen_independent
def test_network_from_dict():
    # Check that a network from a dictionary works
    x = Counter()
    y = Counter()
    d = dict(a=x, b=y)
    net = Network()
    net.add(d)
    net.run(1 * ms)
    assert_equal(len(net.objects), 2)
    assert_equal(x.count, 10)
    assert_equal(y.count, 10)


class NameLister(BrianObject):
    add_to_magic_network = True
    updates = []

    def __init__(self, **kwds):
        super().__init__(**kwds)

    def run(self):
        NameLister.updates.append(self.name)


@pytest.mark.codegen_independent
def test_network_different_clocks():
    NameLister.updates[:] = []
    # Check that a network with two different clocks functions correctly
    x = NameLister(name="x", dt=0.1 * ms, order=0)
    y = NameLister(name="y", dt=1 * ms, order=1)
    net = Network(x, y)
    net.run(100 * second + defaultclock.dt, report="text")
    updates = "".join(NameLister.updates)[2:]  # ignore the first time step
    assert updates == ("xxxxxxxxxxy" * 100000)


@pytest.mark.codegen_independent
def test_network_different_when():
    # Check that a network with different when attributes functions correctly
    NameLister.updates[:] = []
    x = NameLister(name="x", when="start")
    y = NameLister(name="y", when="end")
    net = Network(x, y)
    net.run(0.3 * ms)
    assert_equal("".join(NameLister.updates), "xyxyxy")


@pytest.mark.codegen_independent
def test_network_default_schedule():
    net = Network()
    assert net.schedule == [
        "start",
        "groups",
        "thresholds",
        "synapses",
        "resets",
        "end",
    ]
    # Set the preference and check that the change is taken into account
    prefs.core.network.default_schedule = list(
        reversed(["start", "groups", "thresholds", "synapses", "resets", "end"])
    )
    assert net.schedule == list(
        reversed(["start", "groups", "thresholds", "synapses", "resets", "end"])
    )


@pytest.mark.codegen_independent
def test_network_schedule_change():
    # Check that a changed schedule is taken into account correctly
    NameLister.updates[:] = []
    x = NameLister(name="x", when="thresholds")
    y = NameLister(name="y", when="resets")
    net = Network(x, y)
    net.run(0.3 * ms)
    assert_equal("".join(NameLister.updates), "xyxyxy")
    NameLister.updates[:] = []
    net.schedule = ["start", "groups", "synapses", "resets", "thresholds", "end"]
    net.run(0.3 * ms)
    assert_equal("".join(NameLister.updates), "yxyxyx")


@pytest.mark.codegen_independent
def test_network_before_after_schedule():
    # Test that before... and after... slot names can be used
    NameLister.updates[:] = []
    x = NameLister(name="x", when="before_resets")
    y = NameLister(name="y", when="after_thresholds")
    net = Network(x, y)
    net.schedule = ["thresholds", "resets", "end"]
    net.run(0.3 * ms)
    assert_equal("".join(NameLister.updates), "yxyxyx")


@pytest.mark.codegen_independent
def test_network_custom_slots():
    # Check that custom slots can be inserted into the schedule
    NameLister.updates[:] = []
    x = NameLister(name="x", when="thresholds")
    y = NameLister(name="y", when="in_between")
    z = NameLister(name="z", when="resets")
    net = Network(x, y, z)
    net.schedule = [
        "start",
        "groups",
        "thresholds",
        "in_between",
        "synapses",
        "resets",
        "end",
    ]
    net.run(0.3 * ms)
    assert_equal("".join(NameLister.updates), "xyzxyzxyz")


@pytest.mark.codegen_independent
def test_network_incorrect_schedule():
    # Test that incorrect arguments provided to schedule raise errors
    net = Network()
    # net.schedule = object()
    with pytest.raises(TypeError):
        setattr(net, "schedule", object())
    # net.schedule = 1
    with pytest.raises(TypeError):
        setattr(net, "schedule", 1)
    # net.schedule = {'slot1', 'slot2'}
    with pytest.raises(TypeError):
        setattr(net, "schedule", {"slot1", "slot2"})
    # net.schedule = ['slot', 1]
    with pytest.raises(TypeError):
        setattr(net, "schedule", ["slot", 1])
    # net.schedule = ['start', 'after_start']
    with pytest.raises(ValueError):
        setattr(net, "schedule", ["start", "after_start"])
    # net.schedule = ['before_start', 'start']
    with pytest.raises(ValueError):
        setattr(net, "schedule", ["before_start", "start"])


@pytest.mark.codegen_independent
def test_schedule_warning():
    previous_device = get_device()
    from uuid import uuid4

    # TestDevice1 supports arbitrary schedules, TestDevice2 does not
    class TestDevice1(Device):
        # These functions are needed during the setup of the defaultclock
        def get_value(self, var):
            return np.array([0.0001])

        def add_array(self, var):
            pass

        def init_with_zeros(self, var, dtype):
            pass

        def fill_with_array(self, var, arr):
            pass

    class TestDevice2(TestDevice1):
        def __init__(self):
            super().__init__()
            self.network_schedule = [
                "start",
                "groups",
                "synapses",
                "thresholds",
                "resets",
                "end",
            ]

    # Unique names are important for getting the warnings again for multiple
    # runs of the test suite
    name1 = f"testdevice_{str(uuid4())}"
    name2 = f"testdevice_{str(uuid4())}"
    all_devices[name1] = TestDevice1()
    all_devices[name2] = TestDevice2()

    set_device(name1)
    assert schedule_propagation_offset() == 0 * ms
    net = Network()
    assert schedule_propagation_offset(net) == 0 * ms

    # Any schedule should work
    net.schedule = list(reversed(net.schedule))
    with catch_logs() as l:
        net.run(0 * ms)
        assert len(l) == 0, "did not expect a warning"

    assert schedule_propagation_offset(net) == defaultclock.dt

    set_device(name2)
    assert schedule_propagation_offset() == defaultclock.dt

    # Using the correct schedule should work
    net.schedule = ["start", "groups", "synapses", "thresholds", "resets", "end"]
    with catch_logs() as l:
        net.run(0 * ms)
        assert len(l) == 0, "did not expect a warning"
    assert schedule_propagation_offset(net) == defaultclock.dt

    # Using another (e.g. the default) schedule should raise a warning
    net.schedule = None
    with catch_logs() as l:
        net.run(0 * ms)
        assert len(l) == 1 and l[0][1].endswith("schedule_conflict")
    reset_device(previous_device)


@pytest.mark.codegen_independent
def test_scheduling_summary_magic():
    basename = f"name{str(uuid.uuid4()).replace('-', '_')}"
    group = NeuronGroup(
        10, "dv/dt = -v/(10*ms) : 1", threshold="v>1", reset="v=1", name=basename
    )
    group.run_regularly("v = rand()", dt=defaultclock.dt * 10, when="end")
    state_mon = StateMonitor(group, "v", record=True, name=f"{basename}_sm")
    inactive_state_mon = StateMonitor(
        group, "v", record=True, name=f"{basename}_sm_ia", when="after_end"
    )
    inactive_state_mon.active = False
    summary_before = scheduling_summary()

    assert [entry.name for entry in summary_before.entries] == [
        f"{basename}_sm",
        f"{basename}_stateupdater",
        f"{basename}_spike_thresholder",
        f"{basename}_spike_resetter",
        f"{basename}_run_regularly",
        f"{basename}_sm_ia",
    ]
    assert [entry.when for entry in summary_before.entries] == [
        "start",
        "groups",
        "thresholds",
        "resets",
        "end",
        "after_end",
    ]
    assert [entry.dt for entry in summary_before.entries] == [
        defaultclock.dt,
        defaultclock.dt,
        defaultclock.dt,
        defaultclock.dt,
        defaultclock.dt * 10,
        defaultclock.dt,
    ]
    assert [entry.active for entry in summary_before.entries] == [
        True,
        True,
        True,
        True,
        True,
        False,
    ]
    assert len(str(summary_before))
    assert len(summary_before._repr_html_())
    run(defaultclock.dt)
    summary_after = scheduling_summary()
    assert str(summary_after) == str(summary_before)
    assert summary_after._repr_html_() == summary_before._repr_html_()


@pytest.mark.codegen_independent
def test_scheduling_summary():
    basename = f"name{str(uuid.uuid4()).replace('-', '_')}"
    group = NeuronGroup(
        10, "dv/dt = -v/(10*ms) : 1", threshold="v>1", reset="v=1", name=basename
    )
    group.run_regularly("v = rand()", dt=defaultclock.dt * 10, when="end")
    state_mon = StateMonitor(group, "v", record=True, name=f"{basename}_sm")
    inactive_state_mon = StateMonitor(
        group, "v", record=True, name=f"{basename}_sm_ia", when="after_end"
    )
    inactive_state_mon.active = False

    @network_operation(name=f"{basename}_net_op", when="before_end")
    def foo():
        pass

    net = Network(group, state_mon, inactive_state_mon, foo)
    summary_before = scheduling_summary(net)
    assert [entry.name for entry in summary_before.entries] == [
        f"{basename}_sm",
        f"{basename}_stateupdater",
        f"{basename}_spike_thresholder",
        f"{basename}_spike_resetter",
        f"{basename}_net_op",
        f"{basename}_run_regularly",
        f"{basename}_sm_ia",
    ]
    assert [entry.when for entry in summary_before.entries] == [
        "start",
        "groups",
        "thresholds",
        "resets",
        "before_end",
        "end",
        "after_end",
    ]
    assert [entry.dt for entry in summary_before.entries] == [
        defaultclock.dt,
        defaultclock.dt,
        defaultclock.dt,
        defaultclock.dt,
        defaultclock.dt,
        defaultclock.dt * 10,
        defaultclock.dt,
    ]
    assert [entry.active for entry in summary_before.entries] == [
        True,
        True,
        True,
        True,
        True,
        True,
        False,
    ]
    assert len(str(summary_before))
    assert len(summary_before._repr_html_())
    run(defaultclock.dt)
    summary_after = scheduling_summary(net)
    assert str(summary_after) == str(summary_before)
    assert summary_after._repr_html_() == summary_before._repr_html_()


class Preparer(BrianObject):
    add_to_magic_network = True

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.did_reinit = False
        self.did_pre_run = False
        self.did_post_run = False

    def reinit(self, level=0):
        self.did_reinit = True

    def before_run(self, namespace=None, level=0):
        self.did_pre_run = True

    def after_run(self):
        self.did_post_run = True


@pytest.mark.codegen_independent
def test_magic_network():
    # test that magic network functions correctly
    x = Counter()
    y = Counter()
    run(10 * ms)
    assert_equal(x.count, 100)
    assert_equal(y.count, 100)

    assert len(repr(magic_network))  # very basic test...
    assert len(str(magic_network))  # very basic test...


class Stopper(BrianObject):
    add_to_magic_network = True

    def __init__(self, stoptime, stopfunc, **kwds):
        super().__init__(**kwds)
        self.stoptime = stoptime
        self.stopfunc = stopfunc

    def run(self):
        self.stoptime -= 1
        if self.stoptime <= 0:
            self.stopfunc()


@pytest.mark.codegen_independent
def test_network_stop():
    # test that Network.stop and global stop() work correctly
    net = Network()
    x = Stopper(10, net.stop)
    net.add(x)
    net.run(10 * ms)
    assert_equal(defaultclock.t, 1 * ms)

    x = Stopper(10, stop)
    net = Network(x)
    net.run(10 * ms)
    assert_equal(defaultclock.t, 1 * ms)


@pytest.mark.codegen_independent
def test_network_operations():
    # test NetworkOperation and network_operation
    seq = []

    def f1():
        seq.append("a")

    op1 = NetworkOperation(f1, when="start", order=1)

    @network_operation
    def f2():
        seq.append("b")

    @network_operation(when="end", order=1)
    def f3():
        seq.append("c")

    # In complex frameworks, network operations might be object methods that
    # access some common data
    class Container:
        def __init__(self):
            self.g1_data = "B"
            self.g2_data = "C"

        def g1(self):
            seq.append(self.g1_data)

        def g2(self):
            seq.append(self.g2_data)

    c = Container()
    c_op1 = NetworkOperation(c.g1)
    c_op2 = NetworkOperation(c.g2, when="end", order=1)
    net = Network(op1, f2, f3, c_op1, c_op2)
    net.run(1 * ms)

    assert_equal("".join(seq), "bBacC" * 10)


@pytest.mark.codegen_independent
def test_incorrect_network_operations():
    # Network operations with more than one argument are not allowed
    def func(x, y):
        pass

    class Container:
        def func(self, x, y):
            pass

    c = Container()

    with pytest.raises(TypeError):
        NetworkOperation(func)
    with pytest.raises(TypeError):
        NetworkOperation(c.func)

    # Incorrect use of @network_operation -- it does not work on an instance
    # method
    try:

        class Container:
            @network_operation
            def func(self):
                pass

        raise AssertionError("expected a TypeError")
    except TypeError:
        pass  # this is what we expected


@pytest.mark.codegen_independent
def test_network_operations_name():
    # test NetworkOperation name input
    seq = []

    def f1():
        seq.append("a")

    def f2():
        seq.append("b")

    def x():
        pass

    op = NetworkOperation(lambda: x)
    assert_equal(op.name, "networkoperation")

    op0 = NetworkOperation(lambda: x, name="named_network")
    assert_equal(op0.name, "named_network")

    op1 = NetworkOperation(f1, name="networkoperation_1")
    op2 = NetworkOperation(f1, name="networkoperation_3")
    op3 = NetworkOperation(f2, name="networkoperation_2")

    net = Network(op1, op2, op3)
    net.run(1 * ms)

    assert_equal("".join(seq), "aba" * 10)


@pytest.mark.codegen_independent
def test_network_active_flag():
    # test that the BrianObject.active flag is recognised by Network.run
    x = Counter()
    y = Counter()
    y.active = False
    run(1 * ms)
    assert_equal(x.count, 10)
    assert_equal(y.count, 0)


@pytest.mark.standalone_compatible
@pytest.mark.multiple_runs
def test_spikes_after_deactivating():
    # Make sure that a spike in the last time step gets cleared. See #1319
    always_spike = NeuronGroup(1, "", threshold="True", reset="")
    spike_mon = SpikeMonitor(always_spike)
    run(defaultclock.dt)
    always_spike.active = False
    run(defaultclock.dt)
    device.build(direct_call=False, **device.build_options)
    assert_equal(spike_mon.t[:], [0] * second)


@pytest.mark.codegen_independent
def test_network_t():
    # test that Network.t works as expected
    x = Counter(dt=1 * ms)
    y = Counter(dt=2 * ms)
    net = Network(x, y)
    net.run(4 * ms)
    assert_equal(net.t, 4 * ms)
    net.run(1 * ms)
    assert_equal(net.t, 5 * ms)
    assert_equal(x.count, 5)
    assert_equal(y.count, 3)
    net.run(0.5 * ms)  # should only update x
    assert_equal(net.t, 5.5 * ms)
    assert_equal(x.count, 6)
    assert_equal(y.count, 3)
    net.run(0.5 * ms)  # shouldn't do anything
    assert_equal(net.t, 6 * ms)
    assert_equal(x.count, 6)
    assert_equal(y.count, 3)
    net.run(0.5 * ms)  # should update x and y
    assert_equal(net.t, 6.5 * ms)
    assert_equal(x.count, 7)
    assert_equal(y.count, 4)

    del x, y, net

    # now test with magic run
    x = Counter(dt=1 * ms)
    y = Counter(dt=2 * ms)
    run(4 * ms)
    assert_equal(magic_network.t, 4 * ms)
    assert_equal(x.count, 4)
    assert_equal(y.count, 2)
    run(4 * ms)
    assert_equal(magic_network.t, 8 * ms)
    assert_equal(x.count, 8)
    assert_equal(y.count, 4)
    run(1 * ms)
    assert_equal(magic_network.t, 9 * ms)
    assert_equal(x.count, 9)
    assert_equal(y.count, 5)


@pytest.mark.codegen_independent
def test_incorrect_dt_defaultclock():
    defaultclock.dt = 0.5 * ms
    G = NeuronGroup(1, "dv/dt = -v / (10*ms) : 1")
    net = Network(G)
    net.run(0.5 * ms)
    defaultclock.dt = 1 * ms
    with pytest.raises(ValueError):
        net.run(0 * ms)


@pytest.mark.codegen_independent
def test_incorrect_dt_custom_clock():
    clock = Clock(dt=0.5 * ms)
    G = NeuronGroup(1, "dv/dt = -v / (10*ms) : 1", clock=clock)
    net = Network(G)
    net.run(0.5 * ms)
    clock.dt = 1 * ms
    with pytest.raises(ValueError):
        net.run(0 * ms)


@pytest.mark.codegen_independent
def test_network_remove():
    x = Counter()
    y = Counter()
    net = Network(x, y)
    net.remove(y)
    net.run(1 * ms)
    assert_equal(x.count, 10)
    assert_equal(y.count, 0)
    # the relevance of this test is when we use weakref.proxy objects in
    # Network.objects, we should be able to add and remove these from
    # the Network just as much as the original objects
    # TODO: Does this test make sense now that Network does not store weak
    #       references by default?
    for obj in copy.copy(net.objects):
        net.remove(obj)
    net.run(1 * ms)
    assert_equal(x.count, 10)
    assert_equal(y.count, 0)


@pytest.mark.codegen_independent
def test_contained_objects():
    obj = CounterWithContained()
    net = Network(obj)
    # The contained object should not be stored explicitly
    assert len(net.objects) == 1
    # It should be accessible via the network interface, though
    assert len(net) == 2
    net.run(defaultclock.dt)

    # The contained object should be executed during the run
    assert obj.count == 1
    assert obj.sub_counter.count == 1

    # contained objects should be accessible via get_states/set_states
    states = net.get_states()
    assert len(states) == 2
    assert set(states.keys()) == {obj.name, obj.sub_counter.name}
    assert set(states[obj.name].keys()) == {"state"}
    assert set(states[obj.sub_counter.name].keys()) == {"state"}
    net[obj.name].set_states({"state": 1})
    net[obj.sub_counter.name].set_states({"state": 2})

    net.remove(obj)
    assert len(net.objects) == 0
    assert len(net) == 0
    assert len(net.get_states()) == 0
    net.run(defaultclock.dt)
    assert obj.count == 1
    assert obj.sub_counter.count == 1


class NoninvalidatingCounter(Counter):
    add_to_magic_network = True
    invalidates_magic_network = False


@pytest.mark.codegen_independent
def test_invalid_magic_network():
    x = Counter()
    run(1 * ms)
    assert_equal(x.count, 10)
    y = Counter()
    try:
        run(1 * ms)
        raise AssertionError("Expected a MagicError")
    except MagicError:
        pass  # this is expected
    del x, y
    x = Counter()
    run(1 * ms)
    y = NoninvalidatingCounter()
    run(1 * ms)
    assert_equal(x.count, 20)
    assert_equal(y.count, 10)
    del y
    run(1 * ms)
    assert_equal(x.count, 30)
    del x
    x = Counter()
    run(1 * ms)
    assert_equal(magic_network.t, 1 * ms)
    del x
    x = Counter()
    y = Counter()
    run(1 * ms)
    assert_equal(x.count, 10)
    assert_equal(y.count, 10)


@pytest.mark.codegen_independent
def test_multiple_networks_invalid():
    x = Counter()
    net = Network(x)
    net.run(1 * ms)
    try:
        run(1 * ms)
        raise AssertionError("Expected a RuntimeError")
    except RuntimeError:
        pass  # this is expected

    try:
        net2 = Network(x)
        raise AssertionError("Expected a RuntimeError")
    except RuntimeError:
        pass  # this is expected


@pytest.mark.codegen_independent
def test_magic_weak_reference():
    """
    Test that holding a weak reference to an object does not make it get
    simulated."""

    G1 = NeuronGroup(1, "v:1")

    # this object should not be included
    G2 = weakref.ref(NeuronGroup(1, "v:1"))

    with catch_logs(log_level=logging.DEBUG) as l:
        run(1 * ms)
        # Check the debug messages for the number of included objects
        magic_objects = [
            msg[2] for msg in l if msg[1] == "brian2.core.magic.magic_objects"
        ][0]
        assert "2 objects" in magic_objects, f"Unexpected log message: {magic_objects}"


@pytest.mark.codegen_independent
def test_magic_unused_object():
    """Test that creating unused objects does not affect the magic system."""

    def create_group():
        # Produce two objects but return only one
        G1 = NeuronGroup(1, "v:1")  # no Thresholder or Resetter
        G2 = NeuronGroup(1, "v:1")  # This object should be garbage collected
        return G1

    G = create_group()
    with catch_logs(log_level=logging.DEBUG) as l:
        run(1 * ms)

        # Check the debug messages for the number of included objects
        magic_objects = [
            msg[2] for msg in l if msg[1] == "brian2.core.magic.magic_objects"
        ][0]
        assert "2 objects" in magic_objects, f"Unexpected log message: {magic_objects}"


@pytest.mark.codegen_independent
def test_network_access():
    x = Counter(name="counter")
    net = Network(x)
    assert len(net) == 1
    assert len(repr(net))  # very basic test...
    assert len(str(net))  # very basic test...

    # accessing objects
    assert net["counter"] is x
    with pytest.raises(TypeError):
        net[123]
    with pytest.raises(TypeError):
        net[1:3]
    with pytest.raises(KeyError):
        net["non-existing"]

    objects = [obj for obj in net]
    assert set(objects) == set(net.objects)

    # deleting objects
    del net["counter"]
    with pytest.raises(TypeError):
        net.__delitem__(123)
    with pytest.raises(TypeError):
        net.__delitem__(slice(1, 3))
    with pytest.raises(KeyError):
        net.__delitem__("counter")


@pytest.mark.codegen_independent
def test_dependency_check():
    def create_net():
        G = NeuronGroup(10, "v: 1", threshold="False")
        dependent_objects = [
            StateMonitor(G, "v", record=True),
            SpikeMonitor(G),
            PopulationRateMonitor(G),
            Synapses(G, G, on_pre="v+=1"),
        ]
        return dependent_objects

    dependent_objects = create_net()
    # Trying to simulate the monitors/synapses without the group should fail
    for obj in dependent_objects:
        with pytest.raises(ValueError):
            Network(obj).run(0 * ms)

    # simulation with a magic network should work when we have an explicit
    # reference to one of the objects, but the object should be inactive and
    # we should get a warning
    assert all(obj.active for obj in dependent_objects)
    for obj in dependent_objects:  # obj is our explicit reference
        with catch_logs() as l:
            run(0 * ms)
            dependency_warnings = [
                msg[2] for msg in l if msg[1] == "brian2.core.magic.dependency_warning"
            ]
            assert len(dependency_warnings) == 1
        assert not obj.active


def test_loop():
    """
    Somewhat realistic test with a loop of magic networks
    """

    def run_simulation():
        G = NeuronGroup(10, "dv/dt = -v / (10*ms) : 1", reset="v=0", threshold="v>1")
        G.v = np.linspace(0, 1, 10)
        run(1 * ms)
        # We return potentially problematic references to a VariableView
        return G.v

    # First run
    with catch_logs(log_level=logging.DEBUG) as l:
        v = run_simulation()
        assert v[0] == 0 and 0 < v[-1] < 1
        # Check the debug messages for the number of included objects
        magic_objects = [
            msg[2] for msg in l if msg[1] == "brian2.core.magic.magic_objects"
        ][0]
        assert "4 objects" in magic_objects

    # Second run
    with catch_logs(log_level=logging.DEBUG) as l:
        v = run_simulation()
        assert v[0] == 0 and 0 < v[-1] < 1
        # Check the debug messages for the number of included objects
        magic_objects = [
            msg[2] for msg in l if msg[1] == "brian2.core.magic.magic_objects"
        ][0]
        assert "4 objects" in magic_objects


@pytest.mark.codegen_independent
def test_magic_collect():
    """
    Make sure all expected objects are collected in a magic network
    """
    P = PoissonGroup(10, rates=100 * Hz)
    G = NeuronGroup(10, "v:1", threshold="False")
    S = Synapses(G, G, "")

    state_mon = StateMonitor(G, "v", record=True)
    spike_mon = SpikeMonitor(G)
    rate_mon = PopulationRateMonitor(G)

    objects = collect()

    assert len(objects) == 6, f"expected {int(6)} objects, got {len(objects)}"


import sys
from contextlib import contextmanager
from io import BytesIO, StringIO


@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


@pytest.mark.codegen_independent
def test_progress_report():
    """
    Very basic test of progress reporting
    """
    G = NeuronGroup(1, "")
    net = Network(G)

    # No output
    with captured_output() as (out, err):
        net.run(1 * ms, report=None)
    # There should be at least two lines of output
    out, err = out.getvalue(), err.getvalue()
    assert len(out) == 0 and len(err) == 0

    with captured_output() as (out, err):
        net.run(1 * ms)
    # There should be at least two lines of output
    out, err = out.getvalue(), err.getvalue()
    assert len(out) == 0 and len(err) == 0

    # Progress should go to stdout
    with captured_output() as (out, err):
        net.run(1 * ms, report="text")
    # There should be at least two lines of output
    out, err = out.getvalue(), err.getvalue()
    assert len(out.split("\n")) >= 2 and len(err) == 0

    with captured_output() as (out, err):
        net.run(1 * ms, report="stdout")
    # There should be at least two lines of output
    out, err = out.getvalue(), err.getvalue()
    assert len(out.split("\n")) >= 2 and len(err) == 0

    # Progress should go to stderr
    with captured_output() as (out, err):
        net.run(1 * ms, report="stderr")
    # There should be at least two lines of output
    out, err = out.getvalue(), err.getvalue()
    assert len(err.split("\n")) >= 2 and len(out) == 0

    # Custom function
    calls = []

    def capture_progress(elapsed, complete, start, duration):
        calls.append((elapsed, complete, start, duration))

    with captured_output() as (out, err):
        net.run(1 * ms, report=capture_progress)
    out, err = out.getvalue(), err.getvalue()

    assert len(err) == 0 and len(out) == 0
    # There should be at least a call for the start and the end
    assert len(calls) >= 2 and calls[0][1] == 0.0 and calls[-1][1] == 1.0


@pytest.mark.codegen_independent
def test_progress_report_incorrect():
    """
    Test wrong use of the report option
    """
    G = NeuronGroup(1, "")
    net = Network(G)
    with pytest.raises(ValueError):
        net.run(1 * ms, report="unknown")
    with pytest.raises(TypeError):
        net.run(1 * ms, report=object())


@pytest.mark.standalone_compatible
@pytest.mark.multiple_runs
def test_multiple_runs_report_standalone():
    group = NeuronGroup(1, "dv/dt = 1*Hz : 1")
    run(1 * ms, report="text")
    run(1 * ms)
    device.build(direct_call=False, **device.build_options)


@pytest.mark.standalone_compatible
@pytest.mark.multiple_runs
def test_multiple_runs_report_standalone_2():
    group = NeuronGroup(1, "dv/dt = 1*Hz : 1")
    run(1 * ms)
    run(1 * ms, report="text")
    device.build(direct_call=False, **device.build_options)


@pytest.mark.standalone_compatible
@pytest.mark.multiple_runs
def test_multiple_runs_report_standalone_3():
    group = NeuronGroup(1, "dv/dt = 1*Hz : 1")
    run(1 * ms, report="text")
    run(1 * ms, report="text")
    device.build(direct_call=False, **device.build_options)


# This tests a specific limitation of the C++ standalone mode (cannot mix
# multiple report methods)
@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_multiple_runs_report_standalone_incorrect():
    set_device("cpp_standalone", build_on_run=False)
    group = NeuronGroup(1, "dv/dt = 1*Hz : 1")
    run(1 * ms, report="text")
    with pytest.raises(NotImplementedError):
        run(1 * ms, report="stderr")


@pytest.mark.codegen_independent
def test_store_restore():
    source = NeuronGroup(
        10,
        """dv/dt = rates : 1
                                rates : Hz""",
        threshold="v>1",
        reset="v=0",
    )
    source.rates = "i*100*Hz"
    target = NeuronGroup(10, "v:1")
    synapses = Synapses(source, target, model="w:1", on_pre="v+=w")
    synapses.connect(j="i")
    synapses.w = "i*1.0"
    synapses.delay = "i*ms"
    state_mon = StateMonitor(target, "v", record=True)
    spike_mon = SpikeMonitor(source)
    net = Network(source, target, synapses, state_mon, spike_mon)
    net.store()  # default time slot
    net.run(10 * ms)
    net.store("second")
    net.run(10 * ms)
    v_values = state_mon.v[:, :]
    spike_indices, spike_times = spike_mon.it_
    net.restore()  # Go back to beginning
    assert defaultclock.t == 0 * ms
    assert net.t == 0 * ms
    net.run(20 * ms)
    assert_equal(v_values, state_mon.v[:, :])
    assert_equal(spike_indices, spike_mon.i[:])
    assert_equal(spike_times, spike_mon.t_[:])

    # Go back to middle
    net.restore("second")
    assert defaultclock.t == 10 * ms
    assert net.t == 10 * ms
    net.run(10 * ms)
    assert_equal(v_values, state_mon.v[:, :])
    assert_equal(spike_indices, spike_mon.i[:])
    assert_equal(spike_times, spike_mon.t_[:])

    # Go back again (see github issue #681)
    net.restore("second")
    assert defaultclock.t == 10 * ms
    assert net.t == 10 * ms


@pytest.mark.codegen_independent
def test_store_restore_to_file():
    filename = tempfile.mktemp(suffix="state", prefix="brian_test")
    source = NeuronGroup(
        10,
        """
        dv/dt = rates : 1
        rates : Hz
        """,
        threshold="v>1",
        reset="v=0",
    )
    source.rates = "i*100*Hz"
    target = NeuronGroup(10, "v:1")
    synapses = Synapses(source, target, model="w:1", on_pre="v+=w")
    synapses.connect(j="i")
    synapses.w = "i*1.0"
    synapses.delay = "i*ms"
    state_mon = StateMonitor(target, "v", record=True)
    spike_mon = SpikeMonitor(source)
    net = Network(source, target, synapses, state_mon, spike_mon)
    net.store(filename=filename)  # default time slot
    net.run(10 * ms)
    net.store("second", filename=filename)
    net.run(10 * ms)
    v_values = state_mon.v[:, :]
    spike_indices, spike_times = spike_mon.it_

    net.restore(filename=filename)  # Go back to beginning
    assert defaultclock.t == 0 * ms
    assert net.t == 0 * ms
    net.run(20 * ms)
    assert_equal(v_values, state_mon.v[:, :])
    assert_equal(spike_indices, spike_mon.i[:])
    assert_equal(spike_times, spike_mon.t_[:])

    # Go back to middle
    net.restore("second", filename=filename)
    assert defaultclock.t == 10 * ms
    assert net.t == 10 * ms
    net.run(10 * ms)
    assert_equal(v_values, state_mon.v[:, :])
    assert_equal(spike_indices, spike_mon.i[:])
    assert_equal(spike_times, spike_mon.t_[:])
    try:
        os.remove(filename)
    except OSError:
        pass


@pytest.mark.codegen_independent
def test_store_restore_to_file_new_objects():
    # A more realistic test where the objects are completely re-created
    filename = tempfile.mktemp(suffix="state", prefix="brian_test")

    def create_net():
        # Use a bit of a complicated spike and connection pattern with
        # heterogeneous delays

        # Note: it is important that all objects have the same name, this would
        # be the case if we were running this in a new process but to not rely
        # on garbage collection we will assign explicit names here
        source = SpikeGeneratorGroup(
            5,
            np.arange(5).repeat(3),
            [3, 4, 1, 2, 3, 7, 5, 4, 1, 0, 5, 9, 7, 8, 9] * ms,
            name="source",
        )
        target = NeuronGroup(10, "v:1", name="target")
        synapses = Synapses(source, target, model="w:1", on_pre="v+=w", name="synapses")
        synapses.connect("j>=i")
        synapses.w = "i*1.0 + j*2.0"
        synapses.delay = "(5-i)*ms"
        state_mon = StateMonitor(target, "v", record=True, name="statemonitor")
        input_spikes = SpikeMonitor(source, name="input_spikes")
        net = Network(source, target, synapses, state_mon, input_spikes)
        return net

    net = create_net()
    net.store(filename=filename)  # default time slot
    net.run(5 * ms)
    net.store("second", filename=filename)
    net.run(5 * ms)
    input_spike_indices = np.array(net["input_spikes"].i)
    input_spike_times = Quantity(net["input_spikes"].t, copy=True)
    v_values_full_sim = Quantity(net["statemonitor"].v[:, :], copy=True)

    net = create_net()
    net.restore(filename=filename)  # Go back to beginning
    net.run(10 * ms)
    assert_equal(input_spike_indices, net["input_spikes"].i)
    assert_equal(input_spike_times, net["input_spikes"].t)
    assert_equal(v_values_full_sim, net["statemonitor"].v[:, :])

    net = create_net()
    net.restore("second", filename=filename)  # Go back to middle
    net.run(5 * ms)
    assert_equal(input_spike_indices, net["input_spikes"].i)
    assert_equal(input_spike_times, net["input_spikes"].t)
    assert_equal(v_values_full_sim, net["statemonitor"].v[:, :])

    try:
        os.remove(filename)
    except OSError:
        pass


@pytest.mark.codegen_independent
def test_store_restore_to_file_differing_nets():
    # Check that the store/restore mechanism is not used with differing
    # networks
    filename = tempfile.mktemp(suffix="state", prefix="brian_test")

    source = SpikeGeneratorGroup(
        5, [0, 1, 2, 3, 4], [0, 1, 2, 3, 4] * ms, name="source_1"
    )
    mon = SpikeMonitor(source, name="monitor")
    net = Network(source, mon)
    net.store(filename=filename)

    source_2 = SpikeGeneratorGroup(
        5, [0, 1, 2, 3, 4], [0, 1, 2, 3, 4] * ms, name="source_2"
    )
    mon = SpikeMonitor(source_2, name="monitor")
    net = Network(source_2, mon)
    with pytest.raises(KeyError):
        net.restore(filename=filename)

    net = Network(source)  # Without the monitor
    with pytest.raises(KeyError):
        net.restore(filename=filename)


@pytest.mark.codegen_independent
def test_store_restore_magic():
    source = NeuronGroup(
        10,
        """
        dv/dt = rates : 1
        rates : Hz
        """,
        threshold="v>1",
        reset="v=0",
    )
    source.rates = "i*100*Hz"
    target = NeuronGroup(10, "v:1")
    synapses = Synapses(source, target, model="w:1", on_pre="v+=w")
    synapses.connect(j="i")
    synapses.w = "i*1.0"
    synapses.delay = "i*ms"
    state_mon = StateMonitor(target, "v", record=True)
    spike_mon = SpikeMonitor(source)
    store()  # default time slot
    run(10 * ms)
    store("second")
    run(10 * ms)
    v_values = state_mon.v[:, :]
    spike_indices, spike_times = spike_mon.it_

    restore()  # Go back to beginning
    assert magic_network.t == 0 * ms
    run(20 * ms)
    assert defaultclock.t == 20 * ms
    assert_equal(v_values, state_mon.v[:, :])
    assert_equal(spike_indices, spike_mon.i[:])
    assert_equal(spike_times, spike_mon.t_[:])

    # Go back to middle
    restore("second")
    assert magic_network.t == 10 * ms
    run(10 * ms)
    assert defaultclock.t == 20 * ms
    assert_equal(v_values, state_mon.v[:, :])
    assert_equal(spike_indices, spike_mon.i[:])
    assert_equal(spike_times, spike_mon.t_[:])


@pytest.mark.codegen_independent
def test_store_restore_magic_to_file():
    filename = tempfile.mktemp(suffix="state", prefix="brian_test")
    source = NeuronGroup(
        10,
        """
        dv/dt = rates : 1
        rates : Hz
        """,
        threshold="v>1",
        reset="v=0",
    )
    source.rates = "i*100*Hz"
    target = NeuronGroup(10, "v:1")
    synapses = Synapses(source, target, model="w:1", on_pre="v+=w")
    synapses.connect(j="i")
    synapses.w = "i*1.0"
    synapses.delay = "i*ms"
    state_mon = StateMonitor(target, "v", record=True)
    spike_mon = SpikeMonitor(source)
    store(filename=filename)  # default time slot
    run(10 * ms)
    store("second", filename=filename)
    run(10 * ms)
    v_values = state_mon.v[:, :]
    spike_indices, spike_times = spike_mon.it_

    restore(filename=filename)  # Go back to beginning
    assert magic_network.t == 0 * ms
    run(20 * ms)
    assert defaultclock.t == 20 * ms
    assert_equal(v_values, state_mon.v[:, :])
    assert_equal(spike_indices, spike_mon.i[:])
    assert_equal(spike_times, spike_mon.t_[:])

    # Go back to middle
    restore("second", filename=filename)
    assert magic_network.t == 10 * ms
    run(10 * ms)
    assert defaultclock.t == 20 * ms
    assert_equal(v_values, state_mon.v[:, :])
    assert_equal(spike_indices, spike_mon.i[:])
    assert_equal(spike_times, spike_mon.t_[:])
    try:
        os.remove(filename)
    except OSError:
        pass


@pytest.mark.codegen_independent
def test_store_restore_spikequeue():
    # See github issue #938
    source = SpikeGeneratorGroup(1, [0], [0] * ms)
    target = NeuronGroup(1, "v : 1")
    conn = Synapses(source, target, on_pre="v += 1", delay=2 * defaultclock.dt)
    conn.connect()
    run(defaultclock.dt)  # Spike is not yet delivered
    store()
    run(2 * defaultclock.dt)
    assert target.v[0] == 1
    restore()
    run(2 * defaultclock.dt)
    assert target.v[0] == 1
    restore()
    run(2 * defaultclock.dt)
    assert target.v[0] == 1


@pytest.mark.skipif(
    not isinstance(get_device(), RuntimeDevice),
    reason="Getting/setting random number state only supported for runtime device.",
)
def test_restore_with_random_state():
    group = NeuronGroup(10, "dv/dt = -v/(10*ms) + (10*ms)**-0.5*xi : 1", method="euler")
    group.v = "rand()"
    mon = StateMonitor(group, "v", record=True)
    store()
    run(10 * ms)
    old_v = np.array(group.v)

    restore()  # Random state is not restored
    run(10 * ms)
    assert np.var(old_v - group.v) > 0  # very basic test for difference

    restore(restore_random_state=True)  # Random state is restored
    run(10 * ms)
    assert_equal(old_v, group.v)


@pytest.mark.codegen_independent
def test_store_restore_restore_synapses():
    group = NeuronGroup(10, "x : 1", threshold="False", reset="", name="group")
    synapses = Synapses(group, group, on_pre="x += 1", name="synapses")
    synapses.connect(i=[1, 3, 5], j=[6, 4, 2])
    net = Network(group, synapses)

    tmp_file = tempfile.mktemp()
    net.store(filename=tmp_file)

    # clear up
    del net
    del synapses
    del group

    # Recreate the network without connecting the synapses
    group = NeuronGroup(10, "x: 1", threshold="False", reset="", name="group")
    synapses = Synapses(group, group, "", on_pre="x += 1", name="synapses")
    net = Network(group, synapses)
    try:
        net.restore(filename=tmp_file)

        assert len(synapses) == 3
        assert_array_equal(synapses.i, [1, 3, 5])
        assert_array_equal(synapses.j, [6, 4, 2])

        # Tunning the network should not raise an error, despite the lack
        # of Synapses.connect
        net.run(0 * ms)
    finally:
        os.remove(tmp_file)


@pytest.mark.codegen_independent
def test_defaultclock_dt_changes():
    BrianLogger.suppress_name("resolution_conflict")
    for dt in [0.1 * ms, 0.01 * ms, 0.5 * ms, 1 * ms, 3.3 * ms]:
        defaultclock.dt = dt
        G = NeuronGroup(1, "v:1")
        mon = StateMonitor(G, "v", record=True)
        net = Network(G, mon)
        net.run(2 * dt)
        assert_equal(mon.t[:], [0, dt / ms] * ms)


@pytest.mark.standalone_compatible
@pytest.mark.multiple_runs
def test_dt_changes_between_runs():
    defaultclock.dt = 0.1 * ms
    G = NeuronGroup(1, "v:1")
    mon = StateMonitor(G, "v", record=True)
    run(0.5 * ms)
    defaultclock.dt = 0.5 * ms
    run(0.5 * ms)
    defaultclock.dt = 0.1 * ms
    run(0.5 * ms)
    device.build(direct_call=False, **device.build_options)
    assert len(mon.t[:]) == 5 + 1 + 5
    assert_allclose(
        mon.t[:], [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 1.0, 1.1, 1.2, 1.3, 1.4] * ms
    )


@pytest.mark.codegen_independent
def test_dt_restore():
    defaultclock.dt = 0.5 * ms
    G = NeuronGroup(1, "dv/dt = -v/(10*ms) : 1")
    mon = StateMonitor(G, "v", record=True)
    net = Network(G, mon)
    net.store()

    net.run(1 * ms)
    assert_equal(mon.t[:], [0, 0.5] * ms)
    defaultclock.dt = 1 * ms
    net.run(2 * ms)
    assert_equal(mon.t[:], [0, 0.5, 1, 2] * ms)
    net.restore()
    assert_equal(mon.t[:], [])
    net.run(1 * ms)
    assert defaultclock.dt == 0.5 * ms
    assert_equal(mon.t[:], [0, 0.5] * ms)


@pytest.mark.codegen_independent
def test_continuation():
    defaultclock.dt = 1 * ms
    G = NeuronGroup(1, "dv/dt = -v / (10*ms) : 1")
    G.v = 1
    mon = StateMonitor(G, "v", record=True)
    net = Network(G, mon)
    net.run(2 * ms)

    # Run the same simulation but with two runs that use sub-dt run times
    G2 = NeuronGroup(1, "dv/dt = -v / (10*ms) : 1")
    G2.v = 1
    mon2 = StateMonitor(G2, "v", record=True)
    net2 = Network(G2, mon2)
    net2.run(0.5 * ms)
    net2.run(1.5 * ms)

    assert_equal(mon.t[:], mon2.t[:])
    assert_equal(mon.v[:], mon2.v[:])


@pytest.mark.codegen_independent
def test_get_set_states():
    G = NeuronGroup(10, "v:1", name="a_neurongroup")
    G.v = "i"
    net = Network(G)
    states1 = net.get_states()
    states2 = magic_network.get_states()
    states3 = net.get_states(read_only_variables=False)
    assert (
        set(states1.keys())
        == set(states2.keys())
        == set(states3.keys())
        == {"a_neurongroup"}
    )
    assert (
        set(states1["a_neurongroup"].keys())
        == set(states2["a_neurongroup"].keys())
        == {"i", "dt", "N", "t", "v", "t_in_timesteps"}
    )
    assert set(states3["a_neurongroup"]) == {"v"}

    # Try re-setting the state
    G.v = 0
    net.set_states(states3)
    assert_equal(G.v, np.arange(10))


@pytest.mark.codegen_independent
def test_multiple_runs_defaultclock():
    defaultclock.dt = 0.1 * ms
    G = NeuronGroup(1, "dv/dt = -v / (10*ms) : 1")
    net = Network(G)
    net.run(0.5 * ms)

    # The new dt is not compatible with the previous time but it should not
    # raise an error because we start a new simulation at time 0
    defaultclock.dt = 1 * ms
    G = NeuronGroup(1, "dv/dt = -v / (10*ms) : 1")
    net = Network(G)
    net.run(1 * ms)


@pytest.mark.codegen_independent
def test_multiple_runs_defaultclock_incorrect():
    defaultclock.dt = 0.1 * ms
    G = NeuronGroup(1, "dv/dt = -v / (10*ms) : 1")
    net = Network(G)
    net.run(0.5 * ms)

    # The new dt is not compatible with the previous time since we cannot
    # continue at 0.5ms with a dt of 1ms
    defaultclock.dt = 1 * ms
    with pytest.raises(ValueError):
        net.run(1 * ms)


@pytest.mark.standalone_compatible
def test_profile():
    G = NeuronGroup(
        10,
        "dv/dt = -v / (10*ms) : 1",
        threshold="v>1",
        reset="v=0",
        name="profile_test",
    )
    G.v = 1.1
    net = Network(G)
    net.run(1 * ms, profile=True)
    # The should be four simulated CodeObjects, one for the group and one each
    # for state update, threshold and reset + 1 for the clock
    info = net.profiling_info
    info_dict = dict(info)
    # Standalone does not include the NeuronGroup object (which is not doing
    # anything during the run) in the profiling information, while runtime
    # does
    assert 3 <= len(info) <= 4
    assert len(info) == 3 or "profile_test" in info_dict
    for obj in ["stateupdater", "spike_thresholder", "spike_resetter"]:
        name = f"profile_test_{obj}"
        assert name in info_dict or f"{name}_codeobject" in info_dict
    assert all([t >= 0 * second for _, t in info])


@pytest.mark.standalone_compatible
def test_profile_off():
    G = NeuronGroup(
        10,
        "dv/dt = -v / (10*ms) : 1",
        threshold="v>1",
        reset="v=0",
        name="profile_test",
    )
    net = Network(G)
    net.run(1 * ms, profile=False)
    with pytest.raises(ValueError):
        profiling_summary(net)


@pytest.mark.codegen_independent
def test_profile_ipython_html():
    G = NeuronGroup(
        10,
        "dv/dt = -v / (10*ms) : 1",
        threshold="v>1",
        reset="v=0",
        name="profile_test",
    )
    G.v = 1.1
    net = Network(G)
    net.run(1 * ms, profile=True)
    summary = profiling_summary(net)
    assert len(summary._repr_html_())


@pytest.mark.codegen_independent
def test_magic_scope():
    """
    Check that `start_scope` works as expected.
    """
    G1 = NeuronGroup(1, "v:1", name="G1")
    G2 = NeuronGroup(1, "v:1", name="G2")
    objs1 = {obj.name for obj in collect()}
    start_scope()
    G3 = NeuronGroup(1, "v:1", name="G3")
    G4 = NeuronGroup(1, "v:1", name="G4")
    objs2 = {obj.name for obj in collect()}
    assert objs1 == {"G1", "G2"}
    assert objs2 == {"G3", "G4"}


@pytest.mark.standalone_compatible
def test_runtime_rounding():
    # Test that runtime and standalone round in the same way, see github issue
    # #695 for details
    defaultclock.dt = 20.000000000020002 * us
    G = NeuronGroup(1, "v:1")
    mon = StateMonitor(G, "v", record=True)
    run(defaultclock.dt * 250)
    assert len(mon.t) == 250


@pytest.mark.codegen_independent
def test_small_runs():
    # One long run and multiple small runs should give the same results
    group_1 = NeuronGroup(10, "dv/dt = -v / (10*ms) : 1")
    group_1.v = "(i + 1) / N"
    mon_1 = StateMonitor(group_1, "v", record=True)
    net_1 = Network(group_1, mon_1)
    net_1.run(1 * second)

    group_2 = NeuronGroup(10, "dv/dt = -v / (10*ms) : 1")
    group_2.v = "(i + 1) / N"
    mon_2 = StateMonitor(group_2, "v", record=True)
    net_2 = Network(group_2, mon_2)
    runtime = 1 * ms
    while True:
        runtime *= 3
        runtime = min([runtime, 1 * second - net_2.t])
        net_2.run(runtime)
        if net_2.t >= 1 * second:
            break

    assert_allclose(mon_1.t_[:], mon_2.t_[:])
    assert_allclose(mon_1.v_[:], mon_2.v_[:])


@pytest.mark.codegen_independent
def test_both_equal():
    # check all objects added by Network.add() also have their contained_objects added to 'Network'
    tau = 10 * ms
    diff_eqn = """dv/dt = (1-v)/tau : 1"""
    chg_code = """v = 2*v"""

    Ng = NeuronGroup(1, diff_eqn, method="exact")
    M1 = StateMonitor(Ng, "v", record=True)
    netObj = Network(Ng, M1)
    Ng.run_regularly(chg_code, dt=20 * ms)
    netObj.run(100 * ms)

    start_scope()
    Ng = NeuronGroup(1, diff_eqn, method="exact")
    M2 = StateMonitor(Ng, "v", record=True)
    Ng.run_regularly(chg_code, dt=20 * ms)
    run(100 * ms)

    assert (M1.v == M2.v).all()


@pytest.mark.standalone_compatible
@pytest.mark.multiple_runs
def test_long_run():
    defaultclock.dt = 0.1 * ms
    group = NeuronGroup(1, "x : 1")
    group.run_regularly("x += 1")
    # Timesteps are internally stored as 64bit integers, but previous versions
    # converted them into 32bit integers along the way. We'll make sure that
    # this is not the case and everything runs fine. To not actually run such a
    # long simulation we run a single huge time step
    start_step = 2**31 - 5
    defaultclock.dt = 0.1 * ms
    start_time = start_step * defaultclock.dt
    defaultclock.dt = start_time
    run(start_time)  # A single, *very* long time step
    defaultclock.dt = 0.1 * ms
    run(6 * defaultclock.dt)
    device.build(direct_call=False, **device.build_options)
    assert group.x == 7


@pytest.mark.codegen_independent
def test_long_run_dt_change():
    # Check that the dt check is not too restrictive, see issue #730 for details
    group = NeuronGroup(1, "")  # does nothing...
    defaultclock.dt = 0.1 * ms
    run(100 * second)
    # print profiling_summary()
    defaultclock.dt = 0.01 * ms
    run(1 * second)


@pytest.mark.standalone_compatible
@pytest.mark.multiple_runs
def test_multiple_runs_constant_change():
    const_v = 1
    group = NeuronGroup(1, "v = const_v : 1")
    mon = StateMonitor(group, "v", record=0)
    run(defaultclock.dt)
    const_v = 2
    run(defaultclock.dt)
    device.build(direct_call=False, **device.build_options)
    assert_equal(mon.v[0], [1, 2])


@pytest.mark.standalone_compatible
@pytest.mark.multiple_runs
def test_multiple_runs_function_change():
    inp = TimedArray([1, 2], dt=defaultclock.dt)
    group = NeuronGroup(1, "v = inp(t) : 1")
    mon = StateMonitor(group, "v", record=0)
    run(2 * defaultclock.dt)
    inp = TimedArray([0, 0, 3, 4], dt=defaultclock.dt)
    run(2 * defaultclock.dt)
    device.build(direct_call=False, **device.build_options)
    assert_equal(mon.v[0], [1, 2, 3, 4])


@pytest.mark.codegen_independent
def test_unused_object_warning():
    with catch_logs() as logs:
        # Create a NeuronGroup that is not used in the network
        NeuronGroup(1, "v:1", name="never_used")
        # Make sure that it gets garbage collected
        import gc

        gc.collect()
    assert len(logs) == 1
    assert logs[0][0] == "WARNING"
    assert logs[0][1].endswith("unused_brian_object")
    assert "never_used" in logs[0][2]


@pytest.mark.codegen_independent
def test_negative_duration_in_run():
    G = NeuronGroup(1, "v:1")
    with pytest.raises(ValueError):
        run(-1 * second)


@pytest.mark.codegen_independent
def test_negative_duration_in_net_run():
    G = NeuronGroup(1, "v:1")
    net = Network(G)
    with pytest.raises(ValueError):
        net.run(-1 * second)


if __name__ == "__main__":
    BrianLogger.log_level_warn()
    for t in [
        test_incorrect_network_use,
        test_network_contains,
        test_empty_network,
        test_network_single_object,
        test_network_two_objects,
        test_network_from_dict,
        test_network_different_clocks,
        test_network_different_when,
        test_network_default_schedule,
        test_network_schedule_change,
        test_network_before_after_schedule,
        test_network_custom_slots,
        test_network_incorrect_schedule,
        test_schedule_warning,
        test_scheduling_summary_magic,
        test_scheduling_summary,
        test_magic_network,
        test_network_stop,
        test_network_operations,
        test_incorrect_network_operations,
        test_network_operations_name,
        test_network_active_flag,
        test_network_t,
        test_incorrect_dt_defaultclock,
        test_incorrect_dt_custom_clock,
        test_network_remove,
        test_magic_weak_reference,
        test_magic_unused_object,
        test_invalid_magic_network,
        test_multiple_networks_invalid,
        test_network_access,
        test_loop,
        test_magic_collect,
        test_progress_report,
        test_progress_report_incorrect,
        test_multiple_runs_report_standalone,
        test_multiple_runs_report_standalone_2,
        test_multiple_runs_report_standalone_3,
        test_multiple_runs_report_standalone_incorrect,
        test_store_restore,
        test_store_restore_to_file,
        test_store_restore_to_file_new_objects,
        test_store_restore_to_file_differing_nets,
        test_store_restore_magic,
        test_store_restore_magic_to_file,
        test_store_restore_spikequeue,
        test_store_restore_restore_synapses,
        test_defaultclock_dt_changes,
        test_dt_changes_between_runs,
        test_dt_restore,
        test_continuation,
        test_get_set_states,
        test_multiple_runs_defaultclock,
        test_multiple_runs_defaultclock_incorrect,
        test_profile,
        test_profile_off,
        test_profile_ipython_html,
        test_magic_scope,
        test_runtime_rounding,
        test_small_runs,
        test_both_equal,
        test_long_run,
        test_long_run_dt_change,
        test_multiple_runs_constant_change,
        test_multiple_runs_function_change,
    ]:
        set_device(all_devices["runtime"])
        t()
        reinit_and_delete()



"""
Reproduces Figure 12 (simplified three-compartment model) from the following
paper:

Dendritic Low-Threshold Calcium Currents in Thalamic Relay Cells
Alain Destexhe, Mike Neubig, Daniel Ulrich, John Huguenard
Journal of Neuroscience 15 May 1998, 18 (10) 3574-3588

The original NEURON code is available on ModelDB: https://senselab.med.yale.edu/modeldb/ShowModel.cshtml?model=279

Reference for the original morphology:

Rat VB neuron (thalamocortical cell), given by J. Huguenard, stained with
biocytin and traced by A. Destexhe, December 1992.  The neuron is described
in: J.R. Huguenard & D.A. Prince, A novel T-type current underlies prolonged
calcium-dependent burst firing in GABAergic neurons of rat thalamic reticular
nucleus.  J. Neurosci. 12: 3804-3817, 1992.

Available at NeuroMorpho.org:

http://neuromorpho.org/neuron_info.jsp?neuron_name=tc200
NeuroMorpho.Org ID :NMO_00881

Notes
-----
* Completely removed the "Fast mechanism for submembranal Ca++ concentration
  (cai)" -- it did not affect the results presented here
* Time constants for the I_T current are slightly different from the equations
  given in the paper -- the paper calculation seems to be based on 36 degree
  Celsius but the temperature that is used is 34 degrees.
* To reproduce Figure 12C, the "presence of dendritic shunt conductances" meant
  setting g_L to 0.15 mS/cm^2 for the whole neuron.
* Other small discrepancies with the paper -- values from the NEURON code were
  used whenever different from the values stated in the paper
"""

from brian2 import *
from brian2.units.constants import (zero_celsius, faraday_constant as F,
                                    gas_constant as R)

defaultclock.dt = 0.01*ms

VT = -52*mV
El = -76.5*mV  # from code, text says: -69.85*mV

E_Na = 50*mV
E_K = -100*mV
C_d = 7.954  # dendritic correction factor

T = 34*kelvin + zero_celsius # 34 degC (current-clamp experiments)
tadj_HH = 3.0**((34-36)/10.0)  # temperature adjustment for Na & K (original recordings at 36 degC)
tadj_m_T = 2.5**((34-24)/10.0)
tadj_h_T = 2.5**((34-24)/10.0)

shift_I_T = -1*mV

gamma = F/(R*T)  # R=gas constant, F=Faraday constant
Z_Ca = 2  # Valence of Calcium ions
Ca_i = 240*nM  # intracellular Calcium concentration
Ca_o = 2*mM  # extracellular Calcium concentration

eqs = Equations('''
Im = gl*(El-v) - I_Na - I_K - I_T: amp/meter**2
I_inj : amp (point current)
gl : siemens/meter**2

# HH-type currents for spike initiation
g_Na : siemens/meter**2
g_K : siemens/meter**2
I_Na = g_Na * m**3 * h * (v-E_Na) : amp/meter**2
I_K = g_K * n**4 * (v-E_K) : amp/meter**2
v2 = v - VT : volt  # shifted membrane potential (Traub convention)
dm/dt = (0.32*(mV**-1)*(13.*mV-v2)/
        (exp((13.*mV-v2)/(4.*mV))-1.)*(1-m)-0.28*(mV**-1)*(v2-40.*mV)/
        (exp((v2-40.*mV)/(5.*mV))-1.)*m) / ms * tadj_HH: 1
dn/dt = (0.032*(mV**-1)*(15.*mV-v2)/
        (exp((15.*mV-v2)/(5.*mV))-1.)*(1.-n)-.5*exp((10.*mV-v2)/(40.*mV))*n) / ms * tadj_HH: 1
dh/dt = (0.128*exp((17.*mV-v2)/(18.*mV))*(1.-h)-4./(1+exp((40.*mV-v2)/(5.*mV)))*h) / ms * tadj_HH: 1

# Low-threshold Calcium current (I_T)  -- nonlinear function of voltage
I_T = P_Ca * m_T**2*h_T * G_Ca : amp/meter**2
P_Ca : meter/second  # maximum Permeability to Calcium
G_Ca = Z_Ca**2*F*v*gamma*(Ca_i - Ca_o*exp(-Z_Ca*gamma*v))/(1 - exp(-Z_Ca*gamma*v)) : coulomb/meter**3
dm_T/dt = -(m_T - m_T_inf)/tau_m_T : 1
dh_T/dt = -(h_T - h_T_inf)/tau_h_T : 1
m_T_inf = 1/(1 + exp(-(v/mV + 56)/6.2)) : 1
h_T_inf = 1/(1 + exp((v/mV + 80)/4)) : 1
tau_m_T = (0.612 + 1.0/(exp(-(v/mV + 131)/16.7) + exp((v/mV + 15.8)/18.2))) * ms / tadj_m_T: second
tau_h_T = (int(v<-81*mV) * exp((v/mV + 466)/66.6) +
           int(v>=-81*mV) * (28 + exp(-(v/mV + 21)/10.5))) * ms / tadj_h_T: second
''')

# Simplified three-compartment morphology
morpho = Cylinder(x=[0, 38.42]*um, diameter=26*um)
morpho.dend = Cylinder(x=[0, 12.49]*um, diameter=10.28*um)
morpho.dend.distal = Cylinder(x=[0, 84.67]*um, diameter=8.5*um)
neuron = SpatialNeuron(morpho, eqs, Cm=0.88*uF/cm**2, Ri=173*ohm*cm,
                       method='exponential_euler')

neuron.v = -74*mV
# Only the soma has Na/K channels
neuron.main.g_Na = 100*msiemens/cm**2
neuron.main.g_K = 100*msiemens/cm**2
# Apply the correction factor to the dendrites

neuron.dend.Cm *= C_d
neuron.m_T = 'm_T_inf'
neuron.h_T = 'h_T_inf'

mon = StateMonitor(neuron, ['v'], record=True)

store('initial state')


def do_experiment(currents, somatic_density, dendritic_density,
                  dendritic_conductance=0.0379*msiemens/cm**2,
                  HH_currents=True):
    restore('initial state')
    voltages = []
    neuron.P_Ca = somatic_density
    neuron.dend.distal.P_Ca = dendritic_density * C_d
    # dendritic conductance (shunting conductance used for Fig 12C)
    neuron.gl = dendritic_conductance
    neuron.dend.gl = dendritic_conductance * C_d
    if not HH_currents:
        # Shut off spiking (for Figures 12B and 12C)
        neuron.g_Na = 0*msiemens/cm**2
        neuron.g_K = 0*msiemens/cm**2
    run(180*ms)
    store('before current')
    for current in currents:
        restore('before current')
        neuron.main.I_inj = current
        print('.', end='')
        run(320*ms)
        voltages.append(mon[morpho].v[:])  # somatic voltage
    return voltages


## Run the various variants of the model to reproduce Figure 12
mpl.rcParams['lines.markersize'] = 3.0
fig, axes = plt.subplots(2, 2)
print('Running experiments for Figure A1 ', end='')
voltages = do_experiment([50, 75]*pA, somatic_density=1.7e-5*cm/second,
                         dendritic_density=1.7e-5*cm/second)
print(' done.')
cut_off = 100*ms  # Do not display first part of simulation
axes[0, 0].plot((mon.t - cut_off) / ms, voltages[0] / mV, color='gray')
axes[0, 0].plot((mon.t - cut_off) / ms, voltages[1] / mV, color='black')
axes[0, 0].set(xlim=(0, 400), ylim=(-80, 40), xticks=[],
               title='A1: Uniform T-current density', ylabel='Voltage (mV)')
axes[0, 0].spines['right'].set_visible(False)
axes[0, 0].spines['top'].set_visible(False)
axes[0, 0].spines['bottom'].set_visible(False)

print('Running experiments for Figure A2 ', end='')
voltages = do_experiment([50, 75]*pA, somatic_density=1.7e-5*cm/second,
                         dendritic_density=9.5e-5*cm/second)
print(' done.')
cut_off = 100*ms  # Do not display first part of simulation
axes[1, 0].plot((mon.t - cut_off) / ms, voltages[0] / mV, color='gray')
axes[1, 0].plot((mon.t - cut_off) / ms, voltages[1] / mV, color='black')
axes[1, 0].set(xlim=(0, 400), ylim=(-80, 40),
               title='A2: High T-current density in dendrites',
               xlabel='Time (ms)', ylabel='Voltage (mV)')
axes[1, 0].spines['right'].set_visible(False)
axes[1, 0].spines['top'].set_visible(False)

print('Running experiments for Figure B ', end='')
currents = np.linspace(0, 200, 41)*pA
voltages_somatic = do_experiment(currents, somatic_density=56.36e-5*cm/second,
                                 dendritic_density=0*cm/second,
                                 HH_currents=False)
voltages_somatic_dendritic = do_experiment(currents, somatic_density=1.7e-5*cm/second,
                                           dendritic_density=9.5e-5*cm/second,
                                           HH_currents=False)
print(' done.')
maxima_somatic = Quantity(voltages_somatic).max(axis=1)
maxima_somatic_dendritic = Quantity(voltages_somatic_dendritic).max(axis=1)
axes[0, 1].yaxis.tick_right()
axes[0, 1].plot(currents/pA, maxima_somatic/mV,
                'o-', color='black', label='Somatic only')
axes[0, 1].plot(currents/pA, maxima_somatic_dendritic/mV,
                's-', color='black', label='Somatic & dendritic')
axes[0, 1].set(xlabel='Injected current (pA)', ylabel='Peak LTS (mV)',
               ylim=(-80, 0))
axes[0, 1].legend(loc='best', frameon=False)

print('Running experiments for Figure C ', end='')
currents = np.linspace(200, 400, 41)*pA
voltages_somatic = do_experiment(currents, somatic_density=56.36e-5*cm/second,
                                 dendritic_density=0*cm/second,
                                 dendritic_conductance=0.15*msiemens/cm**2,
                                 HH_currents=False)
voltages_somatic_dendritic = do_experiment(currents, somatic_density=1.7e-5*cm/second,
                                           dendritic_density=9.5e-5*cm/second,
                                           dendritic_conductance=0.15*msiemens/cm**2,
                                           HH_currents=False)
print(' done.')
maxima_somatic = Quantity(voltages_somatic).max(axis=1)
maxima_somatic_dendritic = Quantity(voltages_somatic_dendritic).max(axis=1)
axes[1, 1].yaxis.tick_right()
axes[1, 1].plot(currents/pA, maxima_somatic/mV,
                'o-', color='black', label='Somatic only')
axes[1, 1].plot(currents/pA, maxima_somatic_dendritic/mV,
                's-', color='black', label='Somatic & dendritic')
axes[1, 1].set(xlabel='Injected current (pA)', ylabel='Peak LTS (mV)',
               ylim=(-80, 0))
axes[1, 1].legend(loc='best', frameon=False)

plt.tight_layout()
plt.show()



"""
Implementation of `PoissonGroup`.
"""

import numpy as np

from brian2.core.spikesource import SpikeSource
from brian2.core.variables import Subexpression, Variables
from brian2.groups.group import Group
from brian2.groups.neurongroup import Thresholder
from brian2.groups.subgroup import Subgroup
from brian2.parsing.expressions import parse_expression_dimensions
from brian2.units.fundamentalunits import check_units, fail_for_dimension_mismatch
from brian2.units.stdunits import Hz
from brian2.utils.stringtools import get_identifiers

__all__ = ["PoissonGroup"]


class PoissonGroup(Group, SpikeSource):
    """
    Poisson spike source

    Parameters
    ----------
    N : int
        Number of neurons
    rates : `Quantity`, str
        Single rate, array of rates of length N, or a string expression
        evaluating to a rate. This string expression will be evaluated at every
        time step, it can therefore be time-dependent (e.g. refer to a
        `TimedArray`).
    dt : `Quantity`, optional
        The time step to be used for the simulation. Cannot be combined with
        the `clock` argument.
    clock : `Clock`, optional
        The update clock to be used. If neither a clock, nor the `dt` argument
        is specified, the `defaultclock` will be used.
    when : str, optional
        When to run within a time step, defaults to the ``'thresholds'`` slot.
        See :ref:`scheduling` for possible values.
    order : int, optional
        The priority of of this group for operations occurring at the same time
        step and in the same scheduling slot. Defaults to 0.
    name : str, optional
        Unique name, or use poissongroup, poissongroup_1, etc.
    """

    add_to_magic_network = True

    @check_units(rates=Hz)
    def __init__(
        self,
        N,
        rates,
        dt=None,
        clock=None,
        when="thresholds",
        order=0,
        namespace=None,
        name="poissongroup*",
        codeobj_class=None,
    ):
        Group.__init__(
            self,
            dt=dt,
            clock=clock,
            when=when,
            order=order,
            namespace=namespace,
            name=name,
        )

        self.codeobj_class = codeobj_class

        self._N = N = int(N)

        # TODO: In principle, it would be nice to support Poisson groups with
        # refactoriness, but we can't currently, since the refractoriness
        # information is reset in the state updater which we are not using
        # We could either use a specific template or simply not bother and make
        # users write their own NeuronGroup (with threshold rand() < rates*dt)
        # for more complex use cases.

        self.variables = Variables(self)
        # standard variables
        self.variables.add_constant("N", value=self._N)
        self.variables.add_arange("i", self._N, constant=True, read_only=True)
        self.variables.add_array("_spikespace", size=N + 1, dtype=np.int32)
        self.variables.create_clock_variables(self._clock)

        # The firing rates
        if isinstance(rates, str):
            self.variables.add_subexpression("rates", dimensions=Hz.dim, expr=rates)
        else:
            self.variables.add_array("rates", size=N, dimensions=Hz.dim)
        self._rates = rates

        self.start = 0
        self.stop = N

        self._refractory = False

        self.events = {"spike": "rand() < rates * dt"}
        self.thresholder = {"spike": Thresholder(self)}
        self.contained_objects.append(self.thresholder["spike"])

        self._enable_group_attributes()

        if not isinstance(rates, str):
            self.rates = rates

    def __getitem__(self, item):
        if not isinstance(item, slice):
            raise TypeError("Subgroups can only be constructed using slicing syntax")
        start, stop, step = item.indices(self._N)
        if step != 1:
            raise IndexError("Subgroups have to be contiguous")
        if start >= stop:
            raise IndexError(
                f"Illegal start/end values for subgroup, {int(start)}>={int(stop)}"
            )

        return Subgroup(self, start, stop)

    def before_run(self, run_namespace=None):
        rates_var = self.variables["rates"]
        if isinstance(rates_var, Subexpression):
            # Check that the units of the expression make sense
            expr = rates_var.expr
            identifiers = get_identifiers(expr)
            variables = self.resolve_all(
                identifiers, run_namespace, user_identifiers=identifiers
            )
            unit = parse_expression_dimensions(rates_var.expr, variables)
            fail_for_dimension_mismatch(
                unit,
                Hz,
                "The expression provided for "
                "PoissonGroup's 'rates' "
                "argument, has to have units "
                "of Hz",
            )
        super().before_run(run_namespace)

    @property
    def spikes(self):
        """
        The spikes returned by the most recent thresholding operation.
        """
        # Note that we have to directly access the ArrayVariable object here
        # instead of using the Group mechanism by accessing self._spikespace
        # Using the latter would cut _spikespace to the length of the group
        spikespace = self.variables["_spikespace"].get_value()
        return spikespace[: spikespace[-1]]

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.N}, rates={self._rates!r})"



{# USES_VARIABLES { _queue } #}
{% extends 'common_group.pyx' %}

{% block maincode %}
    cdef _numpy.ndarray[int32_t, ndim=1, mode='c'] _spiking_synapses = _queue.peek()

    # scalar code
    _vectorisation_idx = 1
    {{ scalar_code | autoindent }}

    cdef size_t _spiking_synapse_idx

    for _spiking_synapse_idx in range(len(_spiking_synapses)):
        # vector code
        _idx = _spiking_synapses[_spiking_synapse_idx]
        _vectorisation_idx = _idx
        {{ vector_code | autoindent }}

    # Advance the spike queue
    _queue.advance()

{% endblock %}
